#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
base_obj: base objects with no internal imports for the MMGen suite
"""

class AsyncInit(type):
	async def __call__(cls, *args, **kwargs):
		instance = cls.__new__(cls, *args, **kwargs)
		await type(instance).__init__(instance, *args, **kwargs)
		return instance

class AttrCtrlMeta(type):
	def __call__(cls, *args, **kwargs):
		instance = super().__call__(*args, **kwargs)
		if instance._autolock:
			instance._lock()
		return instance

class AttrCtrl(metaclass=AttrCtrlMeta):
	"""
	After instance is locked, forbid setting any attribute if the attribute is not present
	in either the class or instance dict.

	Ensure that attribute's type matches that of the instance attribute, or the class
	attribute, if _use_class_attr is True.  If the instance or class attribute is set
	to None, no type checking is performed.

	If _default_to_none is True, return None when accessing non-existent attributes
	instead of raising AttributeError.
	"""
	_autolock = True
	_locked = False
	_use_class_attr = False
	_default_to_none = False
	_skip_type_check = ()
	_delete_ok = ()

	def _lock(self):
		self._locked = True

	def __getattr__(self, name):
		if self._locked and self._default_to_none:
			return None
		else:
			raise AttributeError(f'{self} has no attribute {name!r}')

	def __setattr__(self, name, value):

		if self._locked:
			assert name != '_locked', 'lock can be set only once'

			def do_error(name, value, ref_val):
				raise AttributeError(
					f'{value!r}: invalid value for attribute {name!r}'
					+ ' of {} object (must be of type {}, not {})'.format(
						type(self).__name__,
						type(ref_val).__name__,
						type(value).__name__))

			if not (name in self.__dict__ or hasattr(type(self), name)):
				raise AttributeError(f'{type(self).__name__} object has no attribute {name!r}')

			ref_val = getattr(type(self), name) if self._use_class_attr else getattr(self, name)

			if (
				(name not in self._skip_type_check)
				and (ref_val is not None)
				and not isinstance(value, type(ref_val))
			):
				do_error(name, value, ref_val)

		return object.__setattr__(self, name, value)

	def __delattr__(self, name):
		if self._locked and not name in self._delete_ok:
			raise AttributeError('attribute cannot be deleted')
		return object.__delattr__(self, name)

class Lockable(AttrCtrl):
	"""
	After instance is locked, its attributes become read-only, with the following exceptions:
	  - if an attribute's name is in _set_ok, it can be set once after locking, if unset
	  - if an attribute's name is in _reset_ok, read-only restrictions are bypassed and only
	    AttrCtrl checking is performed

	An attribute is considered unset if its value is None, or if it evaluates to False but is
	not zero; or if it is not present in the instance __dict__ when _use_class_attr is True.
	"""
	_set_ok = ()
	_reset_ok = ()

	def _lock(self):
		for name in ('_set_ok', '_reset_ok'):
			for attr in getattr(self, name):
				assert hasattr(self, attr), (
					f'attribute {attr!r} in {name!r} not found in {type(self).__name__} object {id(self)}')
		super()._lock()

	def __setattr__(self, name, value):
		if self._locked and (name in self.__dict__ or hasattr(type(self), name)):
			val = getattr(self, name)
			if name not in (self._set_ok + self._reset_ok):
				raise AttributeError(f'attribute {name!r} of {type(self).__name__} object is read-only')
			elif name not in self._reset_ok:
				if not (
					(val != 0 and not val) or
					(self._use_class_attr and name not in self.__dict__)):
					raise AttributeError(
						f'attribute {name!r} of {type(self).__name__} object is already set,'
						+ ' and resetting is forbidden')

		return AttrCtrl.__setattr__(self, name, value)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.seed: seed file wallet class
"""

from ..util import msg, make_chksum_6, split_into_cols, is_chksum_6
from ..baseconv import baseconv, is_b58_str
from ..seed import Seed
from .unenc import wallet

class wallet(wallet):

	stdin_ok = True
	desc = 'seed data'

	def _format(self):
		b58seed = baseconv('b58').frombytes(self.seed.data, pad='seed', tostr=True)
		self.ssdata.chksum = make_chksum_6(b58seed)
		self.ssdata.b58seed = b58seed
		self.fmt_data = '{} {}\n'.format(
			self.ssdata.chksum,
			split_into_cols(4, b58seed))

	def _deformat(self):
		desc = self.desc
		ld = self.fmt_data.split()

		if not (7 <= len(ld) <= 12): # 6 <= padded b58 data (ld[1:]) <= 11
			msg(f'Invalid data length ({len(ld)}) in {desc}')
			return False

		a, b = ld[0], ''.join(ld[1:])

		if not is_chksum_6(a):
			msg(f'{a!r}: invalid checksum format in {desc}')
			return False

		if not is_b58_str(b):
			msg(f'{b!r}: not a base 58 string, in {desc}')
			return False

		self.cfg._util.vmsg_r(f'Validating {desc} checksum...')

		if not self.cfg._util.compare_chksums(a, 'file', make_chksum_6(b), 'computed', verbose=True):
			return False

		ret = baseconv('b58').tobytes(b, pad='seed')

		if ret is False:
			msg(f'Invalid base-58 encoded seed: {b}')
			return False

		self.seed = Seed(self.cfg, seed_bin=ret)
		self.ssdata.chksum = a
		self.ssdata.b58seed = b

		self.check_usr_seed_len()

		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
examples/whitepaper.py: extract the Bitcoin whitepaper from the blockchain
"""

import asyncio
from mmgen.cfg import Config
from mmgen.rpc import rpc_init

txid = '54e48e5f5c656b26c3bca14a8c95aa583d07ebe84dde3b7dd4a78f4e4186e713'
fn = 'bitcoin.pdf'

async def main():

	cfg = Config(process_opts=True)

	assert cfg.coin == 'BTC' and cfg.network == 'mainnet', 'This script works only on BTC mainnet!'

	c = await rpc_init(cfg, ignore_wallet=True)

	tx = await c.call('getrawtransaction', txid, True)

	chunks = [''.join(d['scriptPubKey']['asm'].split()[1:4]) for d in tx['vout']]

	with open(fn, 'wb') as f:
		f.write(bytes.fromhex(''.join(chunks)[16:368600]))

	print(f'Wrote {fn}')

asyncio.run(main())



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
scripts/exec_wrapper.py: wrapper to launch MMGen scripts in a testing environment
"""

# Import as few modules and define as few names as possible at module level before exec'ing the
# file, as all names will be seen by the exec'ed code.  To prevent name collisions, all names
# defined or imported here at module level should begin with 'exec_wrapper_'

def exec_wrapper_get_colors():
	from collections import namedtuple
	return namedtuple('colors', ['red', 'green', 'yellow', 'blue', 'purple'])(*[
			(lambda s:s) if exec_wrapper_os.getenv('MMGEN_DISABLE_COLOR') else
			(lambda s, n=n:f'\033[{n};1m{s}\033[0m')
		for n in (31, 32, 33, 34, 35)])

def exec_wrapper_init():

	if exec_wrapper_os.path.dirname(exec_wrapper_sys.argv[1]) == 'test':
		# support running of test scripts under wrapper
		repo_root = exec_wrapper_os.getcwd() # assume we’re in repo root
		exec_wrapper_sys.path[0] = repo_root
		# ensure loading of mmgen mods from overlay tree, not repo root:
		from test.overlay import overlay_setup
		overlay_setup(repo_root)
	else:
		exec_wrapper_sys.path.pop(0)

	exec_wrapper_os.environ['MMGEN_EXEC_WRAPPER'] = '1'

def exec_wrapper_write_traceback(e, exit_val):

	import sys, os

	exc_line = (
		f'{type(e).__name__}({e.mmcode}) {e}' if type(e).__name__ in ('MMGenError', 'MMGenSystemExit') else
		f'{type(e).__name__}: {e}')

	c = exec_wrapper_get_colors()

	if os.getenv('EXEC_WRAPPER_TRACEBACK'):
		import traceback
		cwd = os.getcwd()
		sys.path.insert(0, cwd)
		from test.overlay import get_overlay_tree_dir
		overlay_path_pfx = os.path.relpath(get_overlay_tree_dir(cwd)) + '/'

		def fixup_fn(fn_in):
			fn = fn_in.removeprefix(cwd+'/').removeprefix(overlay_path_pfx)
			return fn.removesuffix('_orig.py') + '.py' if fn.endswith('_orig.py') else fn

		def gen_output():
			yield 'Traceback (most recent call last):'
			for e in traceback.extract_tb(sys.exc_info()[2]):
				yield '  File "{f}", line {l}, in {n}\n    {L}'.format(
					f = exec_wrapper_execed_file if e.filename == '<string>' else fixup_fn(e.filename),
					l = '(scrubbed)' if os.getenv('MMGEN_TEST_SUITE_DETERMINISTIC') else e.lineno,
					n = e.name,
					L = e.line or 'N/A')

		tb_lines = list(gen_output())

		if 'SystemExit' in exc_line:
			tb_lines.pop()

		if os.getenv('EXEC_WRAPPER_EXIT_VAL') == str(exit_val):
			sys.stdout.write(c.yellow(exc_line) + '\n')
		else:
			sys.stdout.write('{}\n{}\n'.format(
				c.yellow('\n'.join(tb_lines)),
				c.red(exc_line)))
			print(c.blue('{} script exited with error').format(
				'Test' if os.path.dirname(sys.argv[0]) == 'test' else 'Spawned'))

		with open('test.err', 'w') as fp:
			fp.write('\n'.join(tb_lines + [exc_line]))

	else:
		sys.stdout.write(c.purple((f'NONZERO_EXIT[{exit_val}]: ' if exit_val else '') + exc_line) + '\n')

def exec_wrapper_end_msg():
	if (
		exec_wrapper_os.getenv('EXEC_WRAPPER_DO_RUNTIME_MSG')
		and not exec_wrapper_os.getenv('MMGEN_TEST_SUITE_DETERMINISTIC')):
		c = exec_wrapper_get_colors()
		# write to stdout to ensure script output gets to terminal first
		exec_wrapper_sys.stdout.write(c.blue('Runtime: {:0.5f} secs\n'.format(
			exec_wrapper_time.time() - exec_wrapper_tstart)))

def exec_wrapper_tracemalloc_setup():
	exec_wrapper_os.environ['PYTHONTRACEMALLOC'] = '1'
	import tracemalloc
	tracemalloc.start()
	exec_wrapper_sys.stderr.write("INFO → Appending memory allocation stats to 'tracemalloc.log'\n")

def exec_wrapper_tracemalloc_log():
	import tracemalloc, re
	snapshot = tracemalloc.take_snapshot()
	stats = snapshot.statistics('lineno')
	depth = 100
	col1w = 100
	with open('tracemalloc.log', 'a') as fp:
		fp.write('##### TOP {} {} #####\n'.format(depth, ' '.join(exec_wrapper_sys.argv)))
		for stat in stats[:depth]:
			frame = stat.traceback[0]
			fn = re.sub(r'.*\/site-packages\/|.*\/mmgen\/test\/overlay\/tree\/', '', frame.filename)
			fn = re.sub(r'.*\/mmgen\/test\/', 'test/', fn)
			fp.write('{f:{w}} {s:>8.2f} KiB\n'.format(
				f = f'{fn}:{frame.lineno}:',
				s = stat.size/1024,
				w = col1w))
		fp.write('{f:{w}} {s:8.2f} KiB\n\n'.format(
			f = 'TOTAL {}:'.format(' '.join(exec_wrapper_sys.argv))[:col1w],
			s = sum(stat.size for stat in stats) / 1024,
			w = col1w))

def exec_wrapper_do_exit(e, exit_val):
	if exit_val != 0:
		exec_wrapper_write_traceback(e, exit_val)
	else:
		if exec_wrapper_os.getenv('MMGEN_TRACEMALLOC'):
			exec_wrapper_tracemalloc_log()
		exec_wrapper_end_msg()
	exec_wrapper_sys.exit(exit_val)

import sys as exec_wrapper_sys
import os as exec_wrapper_os
import time as exec_wrapper_time

exec_wrapper_init() # sets sys.path[0] to overlay root

if exec_wrapper_os.getenv('MMGEN_TRACEMALLOC'):
	exec_wrapper_tracemalloc_setup()

# import mmgen mods only after sys.path[0] is set to overlay root!
if exec_wrapper_os.getenv('MMGEN_DEVTOOLS'):
	from mmgen.devinit import init_dev as exec_wrapper_init_dev
	exec_wrapper_init_dev()

exec_wrapper_tstart = exec_wrapper_time.time()

try:
	exec_wrapper_sys.argv.pop(0)
	exec_wrapper_execed_file = exec_wrapper_sys.argv[0]
	with open(exec_wrapper_execed_file) as fp:
		exec(fp.read())
except SystemExit as e:
	exec_wrapper_do_exit(e, e.code)
except Exception as e:
	exec_wrapper_do_exit(
		e, e.mmcode if hasattr(e, 'mmcode') else e.code if hasattr(e, 'code') else 1)

if exec_wrapper_os.getenv('MMGEN_TRACEMALLOC'):
	exec_wrapper_tracemalloc_log()

exec_wrapper_end_msg()



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.eth.rpc.local: Ethereum base protocol local RPC client for the MMGen Project
"""

import re

from ....base_obj import AsyncInit
from ....obj import Int
from ....util import die, fmt, oneshot_warning_group
from ....rpc.local import RPCClient

class daemon_warning(oneshot_warning_group):

	class geth:
		color = 'yellow'
		message = 'Geth has not been tested on mainnet. You may experience problems.'

	class erigon:
		color = 'red'
		message = 'Erigon support is EXPERIMENTAL. Use at your own risk!!!'

class CallSigs:
	pass

class EthereumRPCClient(RPCClient, metaclass=AsyncInit):

	async def __init__(
			self,
			cfg,
			proto,
			*,
			daemon,
			backend,
			ignore_wallet):

		self.proto = proto
		self.daemon = daemon
		self.call_sigs = getattr(CallSigs, daemon.id, None)

		super().__init__(
			cfg  = cfg,
			host = proto.rpc_host or cfg.rpc_host or 'localhost',
			port = daemon.rpc_port)

		await self.set_backend_async(backend)

		vi, bh, ci = await self.gathered_call(None, (
				('web3_clientVersion', ()),
				('eth_getBlockByNumber', ('latest', False)),
				('eth_chainId', ()),
			))

		vip = re.match(self.daemon.version_pat, vi, re.ASCII)
		if not vip:
			die(2, fmt(f"""
			Aborting on daemon mismatch:
			  Requested daemon: {self.daemon.id}
			  Running daemon:   {vi}
			""", strip_char='\t').rstrip())

		self.daemon_version = int('{:d}{:03d}{:03d}'.format(*[int(e) for e in vip.groups()]))
		self.daemon_version_str = '{}.{}.{}'.format(*vip.groups())
		self.daemon_version_info = vi

		self.blockcount = int(bh['number'], 16)
		self.cur_date = int(bh['timestamp'], 16)

		self.caps = ()
		if self.daemon.id in ('parity', 'openethereum'):
			if (await self.call('parity_nodeKind'))['capability'] == 'full':
				self.caps += ('full_node',)
			# parity/openethereum return chainID only for dev chain:
			self.chainID = None if ci is None else Int(ci, base=16)
			self.chain = (await self.call('parity_chain')).replace(' ', '_').replace('_testnet', '')
		elif self.daemon.id in ('geth', 'reth', 'erigon'):
			if self.daemon.network == 'mainnet' and hasattr(daemon_warning, self.daemon.id):
				daemon_warning(self.daemon.id)
			self.caps += ('full_node',)
			self.chainID = Int(ci, base=16)
			self.chain = self.proto.chain_ids[self.chainID]

	def make_host_path(self, wallet):
		return ''

	def get_block_from_minconf(self, minconf):
		assert minconf - 1 <= self.blockcount, (
			f'{minconf}: illegal value for ‘minconf’ (exceeds block count)')
		return (
			'pending' if minconf == 0 else
			'latest' if minconf == 1 else
			hex(self.blockcount - (minconf - 1)))

	rpcmethods = (
		'eth_blockNumber',
		'eth_call',
		# Returns the EIP155 chain ID used for transaction signing at the current best block.
		# Parity: Null is returned if not available, ID not required in transactions
		# Erigon: always returns ID, requires ID in transactions
		'eth_chainId',
		'eth_gasPrice',
		'eth_getBalance',
		'eth_getBlockByNumber',
		'eth_getCode',
		'eth_getTransactionByHash',
		'eth_getTransactionCount',
		'eth_getTransactionReceipt',
		'eth_sendRawTransaction',
		'parity_chain',
		'parity_nodeKind',
		'parity_pendingTransactions',
		'txpool_content', # Geth and friends only
	)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
wallet.dieroll: dieroll wallet class
"""

import time
from ..util import msg, msg_r, die, fmt, remove_whitespace
from ..util2 import block_format
from ..seed import Seed
from ..baseconv import baseconv
from .unenc import wallet

class wallet(wallet):

	stdin_ok = True
	desc = 'base6d die roll seed data'
	conv_cls = baseconv
	wl_id = 'b6d'
	mn_type = 'base6d'
	choose_seedlen_prompt = 'Choose a seed length: 1) 128 bits, 2) 192 bits, 3) 256 bits: '
	choose_seedlen_confirm = 'Seed length of {} bits chosen. OK?'
	user_entropy_prompt = 'Would you like to provide some additional entropy from the keyboard?'
	interactive_input = False

	def _format(self):
		d = baseconv('b6d').frombytes(self.seed.data, pad='seed', tostr=True) + '\n'
		self.fmt_data = block_format(d, gw=5, cols=5)

	def _deformat(self):

		d = remove_whitespace(self.fmt_data)
		bc = baseconv('b6d')
		rmap = bc.seedlen_map_rev

		if not len(d) in rmap:
			die('SeedLengthError', '{!r}: invalid length for {} (must be one of {})'.format(
				len(d),
				self.desc,
				list(rmap)))

		# truncate seed to correct length, discarding high bits
		seed_len = rmap[len(d)]
		seed_bytes = bc.tobytes(d, pad='seed')[-seed_len:]

		if self.interactive_input and self.cfg.usr_randchars:
			from ..ui import keypress_confirm
			if keypress_confirm(self.cfg, self.user_entropy_prompt):
				from ..crypto import Crypto
				seed_bytes = Crypto(self.cfg).add_user_random(
					rand_bytes = seed_bytes,
					desc       = 'gathered from your die rolls')
				self.desc += ' plus user-supplied entropy'

		self.seed = Seed(self.cfg, seed_bin=seed_bytes)

		self.check_usr_seed_len()
		return True

	def _get_data_from_user(self, desc):

		if not self.cfg.stdin_tty:
			from ..ui import get_data_from_user
			return get_data_from_user(self.cfg, desc=desc)

		bc = baseconv('b6d')

		self._print_seed_type()

		if self.cfg.seed_len:
			from ..obj import Int
			msg('Using seed length {} (user-configured)'.format(Int(self.cfg.seed_len).hl()))
			assert self.cfg.seed_len // 8 in bc.seedlen_map, f'{self.cfg.seed_len}: invalid seed length'
			seed_bitlen = self.cfg.seed_len
		else:
			seed_bitlen = self._choose_seedlen([n*8 for n in sorted(bc.seedlen_map)])

		nDierolls = bc.seedlen_map[seed_bitlen // 8]

		message = """
			For a {sb}-bit seed you must roll the die {nd} times.  After each die roll,
			enter the result on the keyboard as a digit.  If you make an invalid entry,
			you'll be prompted to re-enter it.
		"""
		msg('\n'+fmt(message.strip()).format(sb=seed_bitlen, nd=nDierolls)+'\n')

		CUR_HIDE = '\033[?25l'
		CUR_SHOW = '\033[?25h'
		cr = '\n' if self.cfg.test_suite else '\r'
		prompt_fs = f'\b\b\b   {cr}Enter die roll #{{}}: {CUR_SHOW}'
		clear_line = '' if self.cfg.test_suite else '\r' + ' ' * 25
		invalid_msg = CUR_HIDE + cr + 'Invalid entry' + ' ' * 11

		from ..term import get_char
		def get_digit(n):
			p = prompt_fs
			while True:
				time.sleep(self.cfg.short_disp_timeout)
				ch = get_char(p.format(n), num_bytes=1)
				if ch in bc.digits:
					msg_r(CUR_HIDE + ' OK')
					return ch
				else:
					msg_r(invalid_msg)
					p = clear_line + prompt_fs

		dierolls, n = [], 1
		while len(dierolls) < nDierolls:
			dierolls.append(get_digit(n))
			n += 1

		msg('Die rolls successfully entered' + CUR_SHOW)
		self.interactive_input = True

		return ''.join(dierolls)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tw.prune: Tracking wallet pruned listaddresses class for the MMGen suite
"""

from ..util import msg, msg_r, rmsg, ymsg
from ..color import red, green, gray, yellow
from ..obj import ListItemAttr
from .addresses import TwAddresses
from .view import CUR_HOME, ERASE_ALL

class TwAddressesPrune(TwAddresses):

	mod_subpath = 'tw.prune'

	class TwAddress(TwAddresses.TwAddress):
		valid_attrs = TwAddresses.TwAddress.valid_attrs | {'tag'}
		tag = ListItemAttr(bool, typeconv=False, reassign_ok=True)

	async def __init__(self, *args, warn_used=False, **kwargs):
		self.warn_used = warn_used
		await super().__init__(*args, **kwargs)

	def gen_display(self, data, cw, fs, color, fmt_method):

		id_save = data[0].al_id
		yes, no = red('Yes '), green('No  ')

		for n, d in enumerate(data, 1):
			if id_save != d.al_id:
				id_save = d.al_id
				yield ''.ljust(self.term_width)
			yield (
				gray(fmt_method(n, d, cw, fs, False, 'Yes ', 'No  ')) if d.tag else
				fmt_method(n, d, cw, fs, True, yes, no))

	def do_prune(self):

		def gen():
			for n, d in enumerate(self.data, 1):
				if d.tag:
					pruned.append(n)
					if d.amt:
						rmsg(f'Warning: pruned address {d.twmmid.addr} has a balance!')
					elif self.warn_used and d.is_used:
						ymsg(f'Warning: pruned address {d.twmmid.addr} is used!')
				else:
					yield d

		pruned = []
		self.reverse = False
		self.do_sort('twmmid')
		self.data = list(gen())

		return pruned

	class action(TwAddresses.action):

		def get_addrnums(self, parent, desc):
			prompt = f'Enter a range or space-separated list of addresses to {desc}: '
			from ..ui import line_input
			msg('')
			while True:
				reply = line_input(parent.cfg, prompt).strip()
				if reply:
					from ..addrlist import AddrIdxList
					from ..obj import get_obj
					selected = get_obj(AddrIdxList, fmt_str=','.join(reply.split()))
					if selected:
						if selected[-1] <= len(parent.disp_data):
							return selected
						msg(f'Address number must be <= {len(parent.disp_data)}')
				else:
					return []

		def query_user(self, desc, addrnum, e):

			from collections import namedtuple
			md = namedtuple('mdata', ['wmsg', 'prompt'])

			m = {
				'amt': md(
					red('Address #{a} ({b}) has a balance of {c}!'.format(
						a = addrnum,
						b = e.twmmid.addr,
						c = e.amt.hl2(color=False, unit=True))),
					'[p]rune anyway, [P]rune all with balance, [s]kip, [S]kip all with balance: ',
				),
				'used': md(
					yellow('Address #{a} ({b}) is used!'.format(
						a = addrnum,
						b = e.twmmid.addr)),
					'[p]rune anyway, [P]rune all used, [s]kip, [S]kip all used: ',
				),
			}

			from ..term import get_char
			valid_res = 'pPsS'
			msg(m[desc].wmsg)

			while True:
				res = get_char(m[desc].prompt, immed_chars=valid_res)
				if res in valid_res:
					msg('')
					return {
						#     auto,  prune
						'p': (False, True),
						'P': (True,  True),
						's': (False, False),
						'S': (True,  False),
					}[res]
				else:
					msg('\nInvalid keypress')

		async def a_prune(self, parent):

			def do_entry(desc, n, addrnum, e):
				if auto[desc]:
					return False
				else:
					auto[desc], prune = self.query_user(desc, addrnum, e)
					dfl[desc] = auto[desc] and prune
					skip_all_used = auto['used'] and not dfl['used']
					if auto[desc]: # we’ve switched to auto mode, so go back and fix up all previous entries
						for idx in addrnums[:n]:
							e = parent.disp_data[idx-1]
							if skip_all_used and e.is_used:
								e.tag = False
							elif desc == 'amt' and e.amt:
								e.tag = prune
							elif desc == 'used' and (e.is_used and not e.amt):
								e.tag = prune
					# skipping all used addrs implies skipping all addrs with balances
					if skip_all_used:
						auto['amt'] = True
						dfl['amt'] = False
					return prune

			addrnums = self.get_addrnums(parent, 'prune')

			dfl  = {'amt': False, 'used': False}  # default prune policy for given property (has amt, is used)
			auto = {'amt': False,  'used': False} # whether to ask the user, or apply default policy automatically

			for n, addrnum in enumerate(addrnums):
				e = parent.disp_data[addrnum-1]
				if e.amt and not dfl['amt']:
					e.tag = do_entry('amt', n, addrnum, e)
				elif parent.warn_used and (e.is_used and not e.amt) and not dfl['used']:
					e.tag = do_entry('used', n, addrnum, e)
				else:
					e.tag = True

			if parent.scroll:
				msg_r(CUR_HOME + ERASE_ALL)

		async def a_unprune(self, parent):
			for addrnum in self.get_addrnums(parent, 'unprune'):
				parent.disp_data[addrnum-1].tag = False

			if parent.scroll:
				msg_r(CUR_HOME + ERASE_ALL)

		async def a_clear_prune_list(self, parent):
			for d in parent.data:
				d.tag = False



#!/usr/bin/env python3

"""
test.daemontest_d.exec: unit test for the MMGen suite's Daemon class
"""

from subprocess import run, PIPE
from collections import namedtuple

from mmgen.color import orange, red
from mmgen.util import fmt_list, in_nix_environment
from mmgen.daemon import CoinDaemon

from ..include.common import cfg, qmsg, qmsg_r, vmsg, msg, msg_r

def test_flags(coin):
	d = CoinDaemon(cfg, network_id=coin)
	vmsg(f'Available opts:  {fmt_list(d.avail_opts, fmt="bare")}')
	vmsg(f'Available flags: {fmt_list(d.avail_flags, fmt="bare")}')
	vals = namedtuple('vals', ['online', 'no_daemonize', 'keep_cfg_file'])

	def gen():
		for opts, flags, val in (
				(None,                       None,              vals(False, False, False)),
				(None,                       ['keep_cfg_file'], vals(False, False, True)),
				(['online'],                 ['keep_cfg_file'], vals(True, False, True)),
				(['online', 'no_daemonize'], ['keep_cfg_file'], vals(True, True, True)),
			):
			d = CoinDaemon(cfg, network_id=coin, opts=opts, flags=flags)
			assert d.flag.keep_cfg_file == val.keep_cfg_file
			assert d.opt.online == val.online
			assert d.opt.no_daemonize == val.no_daemonize
			d.flag.keep_cfg_file = not val.keep_cfg_file
			d.flag.keep_cfg_file = val.keep_cfg_file
			yield d

	return tuple(gen())

def test_flags_err(ut, d):

	def bad1(): d[0].flag.foo = False
	def bad2(): d[0].opt.foo = False
	def bad3(): d[0].opt.no_daemonize = True
	def bad4(): d[0].flag.keep_cfg_file = 'x'
	def bad5(): d[0].opt.no_daemonize = 'x'
	def bad6(): d[0].flag.keep_cfg_file = False
	def bad7(): d[1].flag.keep_cfg_file = True

	ut.process_bad_data((
		('flag (1)', 'ClassFlagsError', 'unrecognized flag', bad1),
		('opt  (1)', 'ClassFlagsError', 'unrecognized opt',  bad2),
		('opt  (2)', 'AttributeError',  'is read-only',      bad3),
		('flag (2)', 'AssertionError',  'not boolean',       bad4),
		('opt  (3)', 'AttributeError',  'is read-only',      bad5),
		('flag (3)', 'ClassFlagsError', 'not set',           bad6),
		('flag (4)', 'ClassFlagsError', 'already set',       bad7),
	))

class unit_tests:

	win_skip = ('start', 'status', 'stop')
	altcoin_deps = ('flags_eth',)

	def _pre(self):
		self.daemon_ctrl_args = ['btc', 'btc_tn', 'btc_rt'] if cfg.no_altcoin_deps else ['all']

	def _test_cmd(self, args_in, network_ids=[], ok=True):
		args = (
			['python3', f'test/{args_in[0]}-coin-daemons.py']
			+ list(args_in[1:])
			+ (network_ids or self.daemon_ctrl_args))
		vmsg('\n' + orange(f"Running '{' '.join(args)}':"))
		redir = None if cfg.verbose else PIPE
		cp = run(args, stdout=redir, stderr=redir, text=True)
		if cp.returncode != 0:
			if cp.stdout:
				msg(cp.stdout)
			if cp.stderr:
				msg(red(cp.stderr))
			return False
		if ok:
			vmsg('')
			qmsg('OK')
		return True

	def flags(self, name, ut):
		qmsg_r('Testing flags and opts (BTC)...')
		vmsg('')
		daemons = test_flags(coin='btc')
		qmsg('OK')
		qmsg_r('Testing error handling for flags and opts...')
		vmsg('')
		test_flags_err(ut, daemons)
		qmsg('OK')
		return True

	def flags_eth(self, name, ut):
		qmsg_r('Testing flags and opts (ETH)...')
		vmsg('')
		daemons = test_flags(coin='eth')
		qmsg('OK')
		qmsg_r('Testing error handling for flags and opts...')
		vmsg('')
		test_flags_err(ut, daemons)
		qmsg('OK')
		return True

	def avail(self, name, ut):
		qmsg_r('Testing availability of coin daemons...')
		from platform import machine
		test_reth = not (cfg.no_altcoin_deps or cfg.fast)
		test_parity = not (
			cfg.no_altcoin_deps
			or machine() in ('riscv64', 'aarch64', 'armv7l')
			or in_nix_environment())
		ret1 = self._test_cmd(
			['start', '--print-version', '--mainnet-only'],
			network_ids = ['btc'] if cfg.no_altcoin_deps else ['btc', 'ltc', 'bch', 'xmr', 'eth'],
			ok = not (test_reth or test_parity))
		ret2 = self._test_cmd(
			['start', '--print-version', '--mainnet-only', '--daemon-id=reth'],
			network_ids = ['eth'],
			ok = not test_parity) if test_reth else True
		ret3 = self._test_cmd(
			['start', '--print-version', '--mainnet-only'],
			network_ids = ['etc'],
			ok = True) if test_parity else True
		return ret1 and ret2 and ret3

	def versions(self, name, ut):
		qmsg_r('Displaying coin daemon versions...')
		ret1 = self._test_cmd(['start', '--print-version'], ok=False)
		ret2 = self._test_cmd(['start', '--print-version', '--mainnet-only'])
		return ret1 and ret2

	def cmds(self, name, ut):
		qmsg_r('Testing start commands for coin daemons...')
		return self._test_cmd(['start', '--testing'])

	def start(self, name, ut):
		qmsg_r('Starting coin daemons...')
		return self._test_cmd(['start'])

	def status(self, name, ut):
		qmsg_r('Checking status of coin daemons...')
		return self._test_cmd(['start'])

	def stop(self, name, ut):
		qmsg_r('Stopping coin daemons...')
		return self._test_cmd(['stop', '--remove-datadir'])



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.rune.tx.info: THORChain transaction info class
"""

from ....tx.info import TxInfo
from ....color import pink
from ....obj import NonNegativeInt

from ...vm.tx.info import VmTxInfo, mmid_disp

class TxInfo(VmTxInfo, TxInfo):

	def format_body(self, blockcount, nonmm_str, max_mmwid, enl, *, terse, sort):
		tx = self.tx
		t = tx.txobj
		fs = """
			From:      {f}{f_mmid}
			Amount:    {a} {c}
			Gas limit: {G}
			Sequence:  {N}
			Memo:      {m}
		""" if tx.is_swap else """
			From:      {f}{f_mmid}
			To:        {t}{t_mmid}
			Amount:    {a} {c}
			Gas limit: {G}
			Sequence:  {N}
		"""
		return fs.strip().replace('\t', '').format(
			f      = t['from'].hl(0),
			t      = None if tx.is_swap else t['to'].hl(0),
			a      = t['amt'].hl(),
			N      = NonNegativeInt(t['sequence']).hl(),
			m      = pink(tx.swap_memo) if tx.is_swap else None,
			c      = tx.proto.dcoin if tx.outputs else '',
			G      = NonNegativeInt(tx.total_gas).hl(),
			f_mmid = mmid_disp(tx.inputs[0], nonmm_str),
			t_mmid = None if tx.is_swap else mmid_disp(tx.outputs[0], nonmm_str)) + '\n\n'

	def format_abs_fee(self, iwidth, /, *, color=None):
		return self.tx.fee.fmt(iwidth, color=color)

	def format_rel_fee(self):
		return ''

	def format_verbose_footer(self):
		return ''



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.shared: Shared methods for the cmdtest.py test suite
"""

from mmgen.util import get_extension
from mmgen.wallet import get_wallet_cls
from mmgen.addrlist import AddrList
from mmgen.passwdlist import PasswordList

from ..include.common import cfg, cmp_or_die, strip_ansi_escapes, joinpath, silence, end_silence
from .include.common import ref_bw_file, ref_bw_hash_preset, ref_dir

class CmdTestShared:
	'shared methods for the cmdtest.py test suite'

	@property
	def segwit_mmtype(self):
		return ('segwit', 'bech32')[bool(cfg.bech32)] if self.segwit else None

	@property
	def segwit_arg(self):
		return ['--type=' + self.segwit_mmtype] if self.segwit_mmtype else []

	def txcreate_ui_common(
			self,
			t,
			caller             = None,
			menu               = [],
			inputs             = '1',
			file_desc          = 'Unsigned transaction',
			input_sels_prompt  = 'to spend',
			bad_input_sels     = False,
			interactive_fee    = '',
			fee_desc           = 'transaction fee',
			fee_info_pat       = None,
			add_comment        = '',
			view               = 't',
			save               = True,
			return_early       = False,
			tweaks             = [],
			used_chg_addr_resp = None,
			auto_chg_addr      = None):

		txdo = (caller or self.test_name)[:4] == 'txdo'

		expect_pat = r'\[q\]uit menu, .*?:.'
		delete_pat = r'Enter account number .*:.'
		confirm_pat = r'OK\?.*:.'

		if used_chg_addr_resp is not None:
			t.expect('reuse harms your privacy.*:.*', used_chg_addr_resp, regex=True)

		if auto_chg_addr is not None:
			e1 = 'Choose a change address:.*Enter a number> '
			e2 = fr'Using .*{auto_chg_addr}.* as.*address'
			res = t.expect([e1, e2], regex=True)
			if res == 0:
				choice = [s.split(')')[0].lstrip() for s in t.p.match[0].split('\n') if auto_chg_addr in s][0]
				t.send(f'{choice}\n')
				t.expect(e2, regex=True)
			t.send('y')

		pat = expect_pat
		for choice in menu + ['q']:
			t.expect(pat, choice, regex=True)
			if self.proto.base_proto == 'Ethereum':
				pat = confirm_pat if pat == delete_pat else delete_pat if choice == 'D' else expect_pat

		if bad_input_sels:
			for r in ('x', '3-1', '9999'):
				t.expect(input_sels_prompt+': ', r+'\n')

		t.expect(input_sels_prompt+': ', inputs+'\n')

		have_est_fee = t.expect([f'{fee_desc}: ', 'OK? (Y/n): ']) == 1

		if have_est_fee and not interactive_fee:
			t.send('y')
		else:
			if have_est_fee:
				t.send('n')
				t.expect(f'{fee_desc}: ', interactive_fee+'\n')
			else:
				t.send(interactive_fee+'\n')
			if fee_info_pat:
				t.expect(fee_info_pat, regex=True)
			t.expect('OK? (Y/n): ', 'y')

		t.expect('(Y/n): ', '\n') # chg amt OK prompt

		if 'confirm_non_mmgen' in tweaks:
			t.expect('Continue? (Y/n)', '\n')

		if 'confirm_chg_non_mmgen' in tweaks:
			t.expect('to confirm: ', 'YES\n')

		t.do_comment(add_comment)

		if return_early:
			return t

		t.view_tx(view)

		if not txdo:
			t.expect('(y/N): ', ('n', 'y')[save])
			t.written_to_file(file_desc)

		return t

	def txsign_ui_common(
			self,
			t,
			caller      = None,
			view        = 't',
			add_comment = '',
			file_desc   = 'Signed transaction',
			ni          = False,
			save        = True,
			do_passwd   = False,
			passwd      = None,
			has_label   = False):

		txdo = (caller or self.test_name)[:4] == 'txdo'

		if do_passwd and txdo:
			t.passphrase('MMGen wallet', passwd or self.wpasswd)

		if not (ni or txdo):
			t.view_tx(view)
			if do_passwd:
				t.passphrase('MMGen wallet', passwd or self.wpasswd)
			t.do_comment(add_comment, has_label=has_label)
			t.expect('(Y/n): ', ('n', 'y')[save])

		t.written_to_file(file_desc)

		return t

	def txsend_ui_common(
			self,
			t,
			caller       = None,
			view         = 'n',
			add_comment  = '',
			file_desc    = 'Sent transaction',
			confirm_send = True,
			bogus_send   = True,
			test         = False,
			quiet        = False,
			contract_addr = None,
			has_label    = False,
			wait         = False):

		txdo = (caller or self.test_name)[:4] == 'txdo'

		if not txdo:
			t.license() # MMGEN_NO_LICENSE is set, so does nothing
			t.view_tx(view)
			t.do_comment(add_comment, has_label=has_label)

		if not test:
			self._do_confirm_send(t, quiet=quiet, confirm_send=confirm_send)

		if bogus_send:
			txid = ''
			t.expect('BOGUS transaction NOT sent')
		elif test == 'tx_proxy':
			t.expect('can be sent')
			return True
		else:
			m = 'TxID: ' if test else 'Transaction sent: '
			txid = strip_ansi_escapes(t.expect_getend(m))
			assert len(txid) == 64, f'{txid!r}: Incorrect txid length!'

		if not test:
			if wait:
				t.expect('Waiting for first confirmation..')
				while True:
					if t.expect(['.', 'OK']):
						break

			if contract_addr:
				_ = strip_ansi_escapes(t.expect_getend('Contract address: '))
				assert _ == contract_addr, f'Contract address mismatch: {_} != {contract_addr}'
			t.written_to_file(file_desc)

		return txid

	def txbump_ui_common(self, t, *, fee, fee_desc='transaction fee', bad_fee=None):
		t.expect('(Y/n): ', 'n') # network-estimated fee OK?
		if bad_fee:
			t.expect(f'{fee_desc}: ', f'{bad_fee}\n')
		t.expect(f'{fee_desc}: ', f'{fee}\n')
		t.expect('(Y/n): ', 'y') # fee OK?
		t.expect('(Y/n): ', 'y') # signoff
		t.expect('(y/N): ', 'n') # edit comment
		t.expect('(y/N): ', 'y') # save TX?
		t.written_to_file('Fee-bumped transaction')
		return t

	def txsign_end(self, t, tnum=None, has_label=False):
		t.expect('Signing transaction')
		t.do_comment(False, has_label=has_label)
		t.expect(r'Save signed transaction.*?\? \(Y/n\): ', 'y', regex=True)
		t.written_to_file('Signed transaction' + (' #' + tnum if tnum else ''))
		return t

	def txsign(
			self,
			wf,
			txfile,
			save       = True,
			has_label  = False,
			extra_opts = [],
			extra_desc = '',
			view       = 'n',
			dfl_wallet = False):
		opts = extra_opts + ['-d', self.tmpdir, txfile] + ([wf] if wf else [])
		wcls = get_wallet_cls(ext = 'mmdat' if dfl_wallet else get_extension(wf))
		t = self.spawn(
			'mmgen-txsign',
			opts,
			extra_desc,
			no_passthru_opts = ['coin'],
			exit_val = None if save or (wcls.enc and wcls.type != 'brain') else 1)
		t.license()
		t.view_tx(view)
		if wcls.enc and wcls.type != 'brain':
			t.passphrase(wcls.desc, self.wpasswd)
		if save:
			self.txsign_end(t, has_label=has_label)
		else:
			t.do_comment(False, has_label=has_label)
			t.expect('Save signed transaction? (Y/n): ', 'n')
			t.expect('not saved')
		return t

	def ref_brain_chk(self, bw_file=ref_bw_file):
		wf = joinpath(ref_dir, bw_file)
		add_args = [f'-l{self.seed_len}', f'-p{ref_bw_hash_preset}']
		return self.walletchk(wf, add_args=add_args, sid=self.ref_bw_seed_id)

	def walletchk(
			self,
			wf,
			wcls       = None,
			add_args   = [],
			sid        = None,
			extra_desc = '',
			dfl_wallet = False):
		hp = self.hash_preset if hasattr(self, 'hash_preset') else '1'
		wcls = wcls or get_wallet_cls(ext=get_extension(wf))
		t = self.spawn(
				'mmgen-walletchk',
				([] if dfl_wallet else ['-i', wcls.fmt_codes[0]])
				+ self.testnet_opt
				+ add_args + ['-p', hp]
				+ ([wf] if wf else []),
				extra_desc       = extra_desc,
				no_passthru_opts = True)
		if wcls.type != 'incog_hidden':
			t.expect(f"Getting {wcls.desc} from file ‘")
		if wcls.enc and wcls.type != 'brain':
			t.passphrase(wcls.desc, self.wpasswd)
			t.expect(['Passphrase is OK', 'Passphrase.* are correct'], regex=True)
		chksum = t.expect_getend(f'Valid {wcls.desc} for Seed ID ')[:8]
		if sid:
			cmp_or_die(chksum, sid)
		return t

	def addrgen(
			self,
			wf,
			check_ref  = False,
			ftype      = 'addr',
			id_str     = None,
			extra_opts = [],
			mmtype     = None,
			stdout     = False,
			dfl_wallet = False,
			no_passthru_opts = False):
		list_type = ftype[:4]
		passgen = list_type == 'pass'
		if not mmtype and not passgen:
			mmtype = self.segwit_mmtype
		t = self.spawn(
				f'mmgen-{list_type}gen',
				['-d', self.tmpdir] + extra_opts +
				([], ['--type='+str(mmtype)])[bool(mmtype)] +
				([], ['--stdout'])[stdout] +
				([], [wf])[bool(wf)] +
				([], [id_str])[bool(id_str)] +
				[getattr(self, f'{list_type}_idx_list')],
				extra_desc       = f'({mmtype})' if mmtype in ('segwit', 'bech32') else '',
				no_passthru_opts = no_passthru_opts)
		t.license()
		wcls = get_wallet_cls(ext = 'mmdat' if dfl_wallet else get_extension(wf))
		t.passphrase(wcls.desc, self.wpasswd)
		t.expect('Passphrase is OK')
		desc = ('address', 'password')[passgen]
		chksum = strip_ansi_escapes(t.expect_getend(rf'Checksum for {desc} data .*?: ', regex=True))
		if check_ref:
			chksum_chk = (
				self.chk_data[self.test_name] if passgen else
				self.chk_data[self.test_name][self.fork][self.proto.testnet])
			cmp_or_die(chksum, chksum_chk, desc=f'{ftype}list data checksum')
		if passgen:
			t.expect('Encrypt password list? (y/N): ', 'N')
		if stdout:
			t.read()
		else:
			fn = t.written_to_file('Password list' if passgen else 'Addresses')
			cls = PasswordList if passgen else AddrList
			silence()
			al = cls(cfg, self.proto, infile=fn, skip_chksum_msg=True) # read back the file we’ve written
			end_silence()
			cmp_or_die(al.chksum, chksum, desc=f'{ftype}list data checksum from file')
		return t

	def keyaddrgen(self, wf, check_ref=False, extra_opts=[], mmtype=None):
		if not mmtype:
			mmtype = self.segwit_mmtype
		args = ['-d', self.tmpdir, self.usr_rand_arg, wf, self.addr_idx_list]
		t = self.spawn('mmgen-keygen',
				([f'--type={mmtype}'] if mmtype else []) + extra_opts + args,
				extra_desc = f'({mmtype})' if mmtype in ('segwit', 'bech32') else '')
		t.license()
		wcls = get_wallet_cls(ext=get_extension(wf))
		t.passphrase(wcls.desc, self.wpasswd)
		chksum = t.expect_getend(r'Checksum for key-address data .*?: ', regex=True)
		if check_ref:
			chksum_chk = self.chk_data[self.test_name][self.fork][self.proto.testnet]
			cmp_or_die(chksum, chksum_chk, desc='key-address list data checksum')
		t.expect('Encrypt key list? (y/N): ', 'y')
		t.usr_rand(self.usr_rand_chars)
		t.hash_preset('new key-address list', '1')
		t.passphrase_new('new key-address list', self.kapasswd)
		t.written_to_file('Encrypted secret keys')
		return t

	def _do_confirm_send(self, t, quiet=False, confirm_send=True, sure=True):
		if sure:
			t.expect('Are you sure you want to broadcast this')
		m = ('YES, I REALLY WANT TO DO THIS', 'YES')[quiet]
		t.expect(f'{m!r} to confirm: ', ('', m)[confirm_send]+'\n')



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
xmrwallet.ops.create: Monero wallet ops for the MMGen Suite
"""

from pathlib import Path

from ...util import msg, msg_r, gmsg, pp_msg, die
from ...addrlist import ViewKeyAddrList

from .wallet import OpWallet

class OpCreate(OpWallet):
	stem    = 'creat'
	wallet_exists = False
	opts    = ('restore_height',)

	def check_uopts(self):
		if self.cfg.restore_height != 'current':
			if int(self.cfg.restore_height or 0) < 0:
				die(1, f'{self.cfg.restore_height}: invalid value for --restore-height (less than zero)')

	async def process_wallet(self, d, fn, last):
		msg_r('') # for pexpect

		if self.cfg.restore_height == 'current':
			restore_height = self.get_coin_daemon_rpc().call_raw('get_height')['height']
		else:
			restore_height = self.cfg.restore_height

		if self.cfg.watch_only:
			ret = self.c.call(
				'generate_from_keys',
				filename       = fn.name,
				password       = d.wallet_passwd,
				address        = d.addr,
				viewkey        = d.viewkey,
				restore_height = restore_height)
		else:
			from ...xmrseed import xmrseed
			ret = self.c.call(
				'restore_deterministic_wallet',
				filename       = fn.name,
				password       = d.wallet_passwd,
				seed           = xmrseed().fromhex(d.sec.wif, tostr=True),
				restore_height = restore_height,
				language       = 'English')

		pp_msg(ret) if self.cfg.debug else msg(f'  Address: {ret["address"]}')
		return True

class OpCreateOffline(OpCreate):

	def __init__(self, cfg, uarg_tuple):

		super().__init__(cfg, uarg_tuple)

		gmsg('\nCreating viewkey-address file for watch-only wallets')
		vkal = ViewKeyAddrList(
			cfg       = self.cfg,
			proto     = self.proto,
			infile    = None,
			addr_idxs = self.uargs.wallets,
			seed      = self.seed_src.seed,
			skip_chksum_msg = True)
		vkf = vkal.file

		# before writing viewkey-address file, shred any old ones in the directory:
		for f in Path(self.asi.xmr_dir).iterdir():
			if f.name.endswith(vkf.ext):
				from ...fileutil import shred_file
				msg(f'\nShredding old viewkey-address file ‘{f}’')
				shred_file(self.cfg, f, iterations=15)

		vkf.write(outdir=self.asi.xmr_dir)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
tx.tx_proxy: tx proxy classes
"""

from ..color import green, pink, orange
from ..util import msg, msg_r, die
from ..http import HTTPClient

class TxProxyClient(HTTPClient):

	def get_form(self, timeout=None):
		return self.get(path=self.form_path, timeout=timeout)

	def post_form(self, *, data, timeout=None):
		return self.post(path=self.form_path, data=data, timeout=timeout)

	def get_form_element(self, text):
		from lxml import html
		root = html.document_fromstring(text)
		res = [e for e in root.forms if e.attrib.get('action', '').endswith(self.form_path)]
		assert res, 'no matching forms!'
		assert len(res) == 1, 'more than one matching form!'
		return res[0]

	def cache_fn(self, desc, *, extra_desc=None):
		return '{}-{}{}.html'.format(self.name, desc, f'-{extra_desc}' if extra_desc else '')

	def save_response(self, data, desc, *, extra_desc=None):
		from ..fileutil import write_data_to_file
		write_data_to_file(
			self.cfg,
			self.cache_fn(desc, extra_desc=extra_desc),
			data,
			desc = f'{desc} page from {orange(self.host)}')

class BlockchairTxProxyClient(TxProxyClient):

	name = 'blockchair'
	host = 'blockchair.com'
	form_path = '/broadcast'
	assets = {
		'avax': 'avalanche',
		'btc':  'bitcoin',
		'bch':  'bitcoin-cash',
		'bnb':  'bnb',
		'dash': 'dash',
		'doge': 'dogecoin',
		'eth':  'ethereum',
		'etc':  'ethereum-classic',
		'ltc':  'litecoin',
		'zec':  'zcash',
	}
	active_assets = () # tried with ETH, doesn’t work

	def create_post_data(self, *, form_text, coin, tx_hex):

		coin = coin.lower()
		assert coin in self.assets, f'coin {coin} not supported by {self.name}'
		asset = self.assets[coin]

		form = self.get_form_element(form_text)
		data = {}

		e = form.find('.//input')
		assert e.attrib['name'] == '_token', 'input name incorrect!'
		data['_token'] = e.attrib['value']

		e = form.find('.//textarea')
		assert e.attrib['name'] == 'data', 'textarea name incorrect!'
		data['data'] = '0x' + tx_hex

		e = form.find('.//button')
		assert e is not None, 'missing button!'

		e = form.find('.//select')
		assert e.attrib['name'] == 'blockchain', 'select element name incorrect!'

		assets = [f.get('value') for f in e.iter() if f.get('value')]
		assert asset in assets, f'coin {coin} ({asset}) not currently supported by {self.name}'

		data['blockchain'] = asset

		return data

	def get_txid(self, *, result_text):
		msg(f'Response parsing TBD.  Check the cached response at {self.cache_fn("result")}')

class EtherscanTxProxyClient(TxProxyClient):
	name = 'etherscan'
	host = 'etherscan.io'
	form_path = '/pushTx'
	assets = {'eth': 'ethereum'}
	active_assets = ('eth',)

	def create_post_data(self, *, form_text, coin, tx_hex):

		form = self.get_form_element(form_text)
		data = {}

		for e in form.findall('.//input'):
			data[e.attrib['name']] = e.attrib['value']

		if len(data) != 4:
			msg('')
			self.save_response(form_text, 'form')
			die(3, f'{len(data)}: unexpected number of keys in data (expected 4)')

		e = form.find('.//textarea')
		data[e.attrib['name']] = '0x' + tx_hex

		return data

	def get_txid(self, *, result_text):
		import json
		from ..obj import CoinTxID, is_coin_txid
		form = self.get_form_element(result_text)
		json_text = form.find('div/div/div')[1].tail
		txid = None
		try:
			txid = json.loads(json_text)['result'].removeprefix('0x')
		except json.JSONDecodeError:
			msg(json_text)
		except Exception as e:
			msg(f'{type(e).__name__}: {e}')
		return CoinTxID(txid) if is_coin_txid(txid) else False

def send_tx(cfg, txhex):

	c = get_client(cfg)
	msg(f'Using {pink(cfg.tx_proxy.upper())} tx proxy')

	msg_r(f'Retrieving form from {orange(c.host)}...')
	form_text = c.get_form(timeout=180)
	msg('done')

	msg_r('Parsing form...')
	post_data = c.create_post_data(
		form_text = form_text,
		coin      = cfg.coin,
		tx_hex    = txhex)
	msg('done')

	if cfg.test:
		msg(f'Form retrieved from {orange(c.host)} and parsed')
		msg(green('Transaction can be sent'))
		return False

	msg_r('Sending data...')
	result_text = c.post_form(data=post_data, timeout=180)
	msg('done')

	msg_r('Parsing response...')
	txid = c.get_txid(result_text=result_text)
	msg('done')

	msg('Transaction ' + (f'sent: {txid.hl()}' if txid else 'send failed'))
	c.save_response(result_text, 'result', extra_desc=txid)

	return txid

tx_proxies = {
	'blockchair': BlockchairTxProxyClient,
	'etherscan':  EtherscanTxProxyClient
}

def get_client(cfg, *, check_only=False):
	proxy = tx_proxies[cfg.tx_proxy]
	if cfg.coin.lower() in proxy.active_assets:
		return True if check_only else proxy(cfg)
	else:
		die(1, f'Coin {cfg.coin} not supported by TX proxy {pink(proxy.name.upper())}')

def check_client(cfg):
	return get_client(cfg, check_only=True)



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
flags: Class flags and opts for the MMGen suite
"""

from .base_obj import AttrCtrl, Lockable
from .util import fmt_list, die

class ClassFlags(AttrCtrl):
	_name = 'flags'
	_desc = 'flag'
	reserved_attrs = ()

	def __init__(self, parent, arg):
		self._parent = parent
		self._available = getattr(self._parent, 'avail_'+self._name)

		for a in self._available:
			if a.startswith('_'):
				die('ClassFlagsError', f'{a!r}: {self._desc} cannot begin with an underscore')
			for b in self.reserved_attrs:
				if a == b:
					die('ClassFlagsError', f'{a!r}: {b} is a reserved name for {self._desc}')

		if arg:
			assert type(arg) in (list, tuple), f"{arg!r}: {self._name!r} must be list or tuple"
		else:
			arg = []

		for e in arg:
			if e not in self._available:
				self.not_available_error(e)

		for e in self._available:
			setattr(self, e, e in arg)

	def __dir__(self):
		return [k for k in self.__dict__ if not k.startswith('_') and not k in self.reserved_attrs]

	def __str__(self):
		return ' '.join(f'{k}={getattr(self, k)}' for k in dir(self))

	def __setattr__(self, name, val):

		if self._locked:

			if name not in self._available:
				self.not_available_error(name)

			if self._name == 'flags':
				assert isinstance(val, bool), f'{val!r} not boolean'
				old_val = getattr(self, name)
				if val and old_val:
					die('ClassFlagsError', f'{self._desc} {name!r} already set')
				if not val and not old_val:
					die('ClassFlagsError', f'{self._desc} {name!r} not set, so cannot be unset')

		super().__setattr__(name, val)

	def not_available_error(self, name):
		die('ClassFlagsError', '{!r}: unrecognized {} for {}: (available {}: {})'.format(
			name,
			self._desc,
			type(self._parent).__name__,
			self._name,
			fmt_list(self._available, fmt='bare')))

class ClassOpts(ClassFlags, Lockable):
	_name = 'opts'
	_desc = 'opt'



#!/usr/bin/env python3

"""
test.modtest_d.scrypt: password hashing unit test for the MMGen suite
"""

from mmgen.util import msg, msg_r

from ..include.common import cfg, qmsg, vmsg, omsg_r, silence, end_silence

class unit_test:

	def run_test(self, name, ut):
		import time

		msg_r('Testing password hashing...')
		qmsg('')

		from mmgen.crypto import Crypto
		crypto = Crypto(cfg)

		salt = bytes.fromhex('f00f' * 16)

		presets = {
			'1': '64898d01ffcea1252c17411e7bd6c12d00191ee8ddcbf95420e5de8f37a7a1cb',
			'2': '625d0c59fcbcf85eca51b4337d3a4952e554c7c95fe32f4199ad0e9e9370e279',
			'3': 'b30a5cf0e606515be4a983cc4a1a4c21e04806f19aaa0ad354a353c83aecd3ee',
			'4': 'a3a1b99393734510b68adf2c2cfdc627a2bc3281913d8ea6fbb677d39781a9fa',
			'5': '0c7e1a672738cee49cf0ff6f3208190ca418e741835fd6995ce9558cc19f3f04',
			'6': '91f9d1c9baf3948433dab58dcc912d96035392c1db21ede96d2f369e025ab06d',
			'7': 'fcb2cd05268de43b0d2d45f78a56e5d446b0bd2d3b57bdbc77cc17a42942f1bd'
		}

		pws = (
			('',      'cc8d99ce7365d8a9d2422d71ce330e130b2cade46a8cc0459a3f83e1a6ac3d30'),
			('foo',   'f0e2cce1d9980edf2373a2070ad3560c2506faf9bc50704a1bc5cdb3c7f63f3b'),
			('φυβαρ', '64898d01ffcea1252c17411e7bd6c12d00191ee8ddcbf95420e5de8f37a7a1cb')
		)

		def test_passwords():
			for pw_base, res in pws:
				for pw in (pw_base, pw_base.encode()):
					pw_disp = "'"+pw+"'" if isinstance(pw, str) else "b'"+pw.decode()+"'"
					if cfg.quiet:
						omsg_r('.')
					else:
						msg_r(f'\n  password {pw_disp:9} ')
					ret = crypto.scrypt_hash_passphrase(pw, salt, '1').hex()
					assert ret == res, ret

		def test_presets(do_presets):
			for hp in do_presets:
				hp = str(hp)
				res = presets[hp]
				pw = 'φυβαρ'
				if cfg.quiet:
					omsg_r('.')
				else:
					msg_r(f'\n  {hp!r:3}: {crypto.hash_presets[hp]!r:12}  ')
				st = time.time()
				ret = crypto.scrypt_hash_passphrase(pw, salt, hp).hex()
				t = time.time() - st
				vmsg('' if cfg.test_suite_deterministic else f'  {t:0.4f} secs')
				assert ret == res, ret

		if cfg.quiet:
			silence()

		cfg.force_standalone_scrypt_module = False
		vmsg('Passwords (auto module selection):')
		test_passwords()
		vmsg('Hash presets (auto module selection):')
		test_presets((1, 2, 3, 4) if cfg.fast else (1, 2, 3, 4, 5, 6, 7))

		cfg.force_standalone_scrypt_module = True
		vmsg('Passwords (force standalone scrypt module):')
		test_passwords()
		vmsg('Hash presets (force standalone scrypt module):')
		test_presets((1, 2, 3))

		if cfg.quiet:
			end_silence()

		msg('OK')
		return True



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
# Licensed under the GNU General Public License, Version 3:
#   https://www.gnu.org/licenses
# Public project repositories:
#   https://github.com/mmgen/mmgen-wallet
#   https://gitlab.com/mmgen/mmgen-wallet

"""
proto.vm.tx.new_swap: new swap transaction methods for VM chains
"""

class VmNewSwap:

	def update_data_output(self, trade_limit):
		parsed_memo = self.swap_proto_mod.Memo.parse(self.swap_memo)
		self.swap_memo = str(self.swap_proto_mod.Memo(
			self.swap_cfg,
			self.recv_proto,
			self.recv_asset,
			self.recv_proto.coin_addr(parsed_memo.address),
			trade_limit = trade_limit))
		self.set_gas_with_data(self.swap_memo.encode())

	@property
	def vault_idx(self):
		return 0

	@property
	def vault_output(self):
		return self.outputs[0]



#!/usr/bin/env python3
#
# MMGen Wallet, a terminal-based cryptocurrency wallet
# Copyright (C)2013-2025 The MMGen Project <mmgen@tuta.io>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
test.cmdtest_d.main: Basic operations tests for the cmdtest.py test suite
"""

import sys, os

from mmgen.util import msg, msg_r, async_run, capfirst, get_extension, die
from mmgen.color import green, cyan, gray
from mmgen.fileutil import get_data_from_file, write_data_to_file
from mmgen.wallet import get_wallet_cls
from mmgen.wallet.mmgen import wallet as MMGenWallet
from mmgen.wallet.incog import wallet as IncogWallet
from mmgen.rpc import rpc_init

from ..include.common import (
	vmsg,
	joinpath,
	silence,
	end_silence,
	getrand,
	getrandnum,
	getrandnum_range,
	getrandhex,
	strip_ansi_escapes
)
from .include.common import (
	pwfile,
	hincog_fn,
	get_file_with_ext,
	get_comment,
	tx_comment_lat_cyr_gr,
	hincog_offset,
	hincog_bytes,
	hincog_seedlen,
	incog_id_fn,
	non_mmgen_fn
)
from .base import CmdTestBase
from .shared import CmdTestShared

def make_brainwallet_file(cfg, fn):
	# Print random words with random whitespace in between
	wl = rwords.split()
	nwords, ws_list, max_spaces = 10, '    \n', 5
	def rand_ws_seq():
		nchars = getrandnum(1) % max_spaces + 1
		return ''.join([ws_list[getrandnum_range(1, 200) % len(ws_list)] for i in range(nchars)])
	rand_pairs = [wl[getrandnum_range(1, 200) % len(wl)] + rand_ws_seq() for i in range(nwords)]
	d = ''.join(rand_pairs).rstrip() + '\n'
	if cfg.verbose:
		msg_r(f'Brainwallet password:\n{cyan(d)}')
	write_data_to_file(cfg, fn, d, desc='brainwallet password', quiet=True, ignore_opt_outdir=True)

def verify_checksum_or_exit(checksum, chk):
	chk = strip_ansi_escapes(chk)
	if checksum != chk:
		die('TestSuiteFatalException', f'Checksum error: {chk}')
	vmsg(green('Checksums match: ') + cyan(chk))

addrs_per_wallet = 8

# 100 words chosen randomly from here:
#   https://github.com/bitcoin/bips/pull/432/files/6332230d63149a950d05db78964a03bfd344e6b0
rwords = """
	алфавит алый амнезия амфора артист баян белый биатлон брат бульвар веревка вернуть весть возраст
	восток горло горный десяток дятел ежевика жест жизнь жрать заговор здание зона изделие итог кабина
	кавалер каждый канал керосин класс клятва князь кривой крыша крючок кузнец кукла ландшафт мальчик
	масса масштаб матрос мрак муравей мычать негодяй носок ночной нрав оборот оружие открытие оттенок
	палуба пароход период пехота печать письмо позор полтора понятие поцелуй почему приступ пруд пятно
	ранее режим речь роса рынок рябой седой сердце сквозь смех снимок сойти соперник спичка стон
	сувенир сугроб суть сцена театр тираж толк удивить улыбка фирма читатель эстония эстрада юность
	"""

class CmdTestMain(CmdTestBase, CmdTestShared):
	'basic operations with emulated tracking wallet'
	tmpdir_nums = [1, 2, 3, 4, 5, 14, 15, 16, 20, 21]
	networks = ('btc', 'btc_tn', 'ltc', 'ltc_tn', 'bch', 'bch_tn')
	passthru_opts = ('daemon_data_dir', 'rpc_port', 'coin', 'testnet', 'rpc_backend')
	segwit_opts_ok = True
	color = True
	need_daemon = True
	cmd_group = (
		('walletgen_dfl_wallet',
			(15, 'wallet generation (default wallet)', [[[], 15]])
		),
		('subwalletgen_dfl_wallet',
			(15, 'subwallet generation (default wallet)', [[[pwfile], 15]])
		),
		('export_seed_dfl_wallet',
			(15, 'seed export to mmseed format (default wallet)', [[[pwfile], 15]])),
		('addrgen_dfl_wallet',
			(15, 'address generation (default wallet)', [[[pwfile], 15]])
		),
		('txcreate_dfl_wallet',
			(15, 'transaction creation (default wallet)', [[['addrs'], 15]])
		),
		('txsign_dfl_wallet',
			(15, 'transaction signing (default wallet)', [[['rawtx', pwfile], 15]])
		),
		('passchg_dfl_wallet',
			(16, 'password, label and hash preset change (default wallet)', [[[pwfile], 15]])
		),
		('walletchk_newpass_dfl_wallet',
			(16, 'wallet check with new pw, label and hash preset', [[[pwfile], 16]])
		),
		('delete_dfl_wallet',
			(15, 'delete default wallet', [[[pwfile], 15]])
		),
		('walletgen',
			(1, 'wallet generation', [[['del_dw_run'], 15]])
		),
		('subwalletgen',
			(1, 'subwallet generation', [[['mmdat'], 1]])
		),
		('subwalletgen_mnemonic',
			(1, 'subwallet generation (to mnemonic format)', [[['mmdat'], 1]])
		),
		# ('walletchk', (1, 'wallet check', [[['mmdat'], 1]])),
		('passchg',
			(5, 'password, label and hash preset change', [[['mmdat', pwfile], 1]])
		),
		('passchg_keeplabel',
			(5, 'password, label and hash preset change (keep label)', [[['mmdat', pwfile], 1]])
		),
		('passchg_usrlabel',
			(5, 'password, label and hash preset change (interactive label)', [[['mmdat', pwfile], 1]])
		),
		('walletchk_newpass',
			(5, 'wallet check with new pw, label and hash preset', [[['mmdat', pwfile], 5]])
		),
		('addrgen',
			(1, 'address generation', [[['mmdat'], 1]])
		),
		('txcreate',
			(1, 'transaction creation', [[['addrs'], 1]])
		),
		('txbump',
			(1, 'transaction fee bumping (no send)', [[['rawtx'], 1]])
		),
		('txsign',
			(1, 'transaction signing', [[['mmdat', 'rawtx'], 1]])
		),
		('txsend',
			(1, 'transaction sending', [[['sigtx'], 1]])
		),
		# txdo must go after txsign
		('txdo',
			(1, 'online transaction', [[['sigtx', 'mmdat'], 1]])
		),
		('export_seed',
			(1, 'seed export to mmseed format', [[['mmdat'], 1]])
		),
		('export_hex',
			(1, 'seed export to hexadecimal format', [[['mmdat'], 1]])
		),
		('export_mnemonic',
			(1, 'seed export to mmwords format', [[['mmdat'], 1]])
		),
		('export_bip39',
			(1, 'seed export to bip39 format', [[['mmdat'], 1]])
		),
		('export_incog',
			(1, 'seed export to mmincog format', [[['mmdat'], 1]])
		),
		('export_incog_hex',
			(1, 'seed export to mmincog hex format', [[['mmdat'], 1]])
		),
		('export_incog_hidden',
			(1, 'seed export to hidden mmincog format', [[['mmdat'], 1]])
		),
		('addrgen_seed',
			(1, 'address generation from mmseed file', [[['mmseed', 'addrs'], 1]])
		),
		('addrgen_hex',
			(1, 'address generation from mmhex file', [[['mmhex', 'addrs'], 1]])
		),
		('addrgen_mnemonic',
			(1, 'address generation from mmwords file', [[['mmwords', 'addrs'], 1]])
		),
		('addrgen_incog',
			(1, 'address generation from mmincog file', [[['mmincog', 'addrs'], 1]])
		),
		('addrgen_incog_hex',
			(1, 'address generation from mmincog hex file', [[['mmincox', 'addrs'], 1]])
		),
		('addrgen_incog_hidden',
			(1, 'address generation from hidden mmincog file', [[[hincog_fn, 'addrs'], 1]])
		),
		('keyaddrgen',
			(1, 'key-address file generation', [[['mmdat'], 1]])
		),
		('txsign_keyaddr',
			(1, 'transaction signing with key-address file', [[['akeys.mmenc', 'rawtx'], 1]])
		),
		('txcreate_ni',
			(1, 'transaction creation (non-interactive)', [[['addrs'], 1]])
		),
		('walletgen2',
			(2, 'wallet generation (2), 128-bit seed', [[['del_dw_run'], 15]])
		),
		('addrgen2',
			(2, 'address generation (2)', [[['mmdat'], 2]])
		),
		('txcreate2',
			(2, 'transaction creation (2)', [[['addrs'], 2]])
		),
		('txsign2',
			(2, 'transaction signing, two transactions', [[['mmdat', 'rawtx'], 1], [['mmdat', 'rawtx'], 2]])
		),
		('export_mnemonic2',
			(2, 'seed export to mmwords format (2)', [[['mmdat'], 2]])
		),
		('walletgen3',
			(3, 'wallet generation (3)', [[['del_dw_run'], 15]])
		),
		('addrgen3',
			(3, 'address generation (3)', [[['mmdat'], 3]])
		),
		('txcreate3',
			(3, 'tx creation with inputs and outputs from two wallets', [[['addrs'], 1], [['addrs'], 3]])
			),
		('txsign3',
			(3, 'tx signing with inputs and outputs from two wallets', [[['mmdat'], 1], [['mmdat', 'rawtx'], 3]])
			),
		('walletgen14',
			(14, 'wallet generation (14)', [[['del_dw_run'], 15]], 14)
		),
		('addrgen14',
			(14, 'address generation (14)', [[['mmdat'], 14]])
		),
		('keyaddrgen14',
			(14, 'key-address file generation (14)', [[['mmdat'], 14]], 14)
		),
		('walletgen4',
			(4, 'wallet generation (4) (brainwallet)', [[['del_dw_run'], 15]])
		),
		('addrgen4',
			(4, 'address generation (4)', [[['mmdat'], 4]])
		),
		('txcreate4', (
			4,
			'tx creation with inputs and outputs from four seed sources, key-address file '
			'and non-MMGen inputs and outputs',
			[
				[['addrs'], 1],
				[['addrs'], 2],
				[['addrs'], 3],
				[['addrs'], 4],
				[['addrs', 'akeys.mmenc'], 14]
			])
		),
		('txsign4', (
			4,
			'tx signing with inputs and outputs from incog file, mnemonic file, wallet, '
			'brainwallet, key-address file and non-MMGen inputs and outputs',
			[
				[['mmincog'], 1],
				[['mmwords'], 2],
				[['mmdat'], 3],
				[['mmbrain', 'rawtx'], 4],
				[['akeys.mmenc'], 14]
			])
		),
		('txdo4', (
			4,
			'tx creation, signing and sending with inputs and outputs from four seed sources, '
			'key-address file and non-MMGen inputs and outputs',
			[
				[['addrs'], 1],
				[['addrs'], 2],
				[['addrs'], 3],
				[['addrs'], 4],
				[['addrs', 'akeys.mmenc'], 14],
				[['mmincog'], 1],
				[['mmwords'], 2],
				[['mmdat'], 3],
				[['mmbrain', 'rawtx'], 4],
				[['akeys.mmenc'], 14]
			])
		), # must go after txsign4
		('txbump4', (
			4,
			'tx fee bump + send with inputs and outputs from four seed sources, key-address file '
			'and non-MMGen inputs and outputs',
			[
				[['akeys.mmenc'], 14],
				[['mmincog'], 1],
				[['mmwords'], 2],
				[['mmdat'], 3],
				[['akeys.mmenc'], 14],
				[['mmbrain', 'sigtx', 'mmdat', 'txdo'], 4]
			])
		), # must go after txsign4
		('walletgen5',
			(20, 'wallet generation (5)', [[['del_dw_run'], 15]], 20)
		),
		('addrgen5',
			(20, 'address generation (5)', [[['mmdat'], 20]])
		),
		('txcreate5',
			(20, 'transaction creation with bad vsize (5)', [[['addrs'], 20]])
		),
		('txsign5',
			(20, 'transaction signing with bad vsize', [[['mmdat', 'rawtx'], 20]])
		),
		('walletgen6',
			(21, 'wallet generation (6)', [[['del_dw_run'], 15]], 21)
		),
		('addrgen6',
			(21, 'address generation (6)', [[['mmdat'], 21]])
		),
		('txcreate6',
			(21, 'transaction creation with corrected vsize (6)', [[['addrs'], 21]])
		),
		('txsign6',
			(21, 'transaction signing with corrected vsize', [[['mmdat', 'rawtx'], 21]])
		),
	)
	segwit_do = (
		'walletgen',
		'addrgen',
		'txcreate',
		'txbump',
		'txsign',
		'txsend',
		'txdo',
		'export_incog',
		'keyaddrgen',
		'txsign_keyaddr',
		'txcreate_ni',
		'walletgen2',
		'addrgen2',
		'txcreate2',
		'txsign2',
		'export_mnemonic2',
		'walletgen3',
		'addrgen3',
		'txcreate3',
		'txsign3',
		'walletgen14',
		'addrgen14',
		'keyaddrgen14',
		'walletgen4',
		'addrgen4',
		'txcreate4',
		'txsign4',
		'txdo4',
		'txbump4',
		'walletgen5',
		'addrgen5',
		'txcreate5',
		'txsign5',
		'walletgen6',
		'addrgen6',
		'txcreate6',
		'txsign6',
	)

	def __init__(self, cfg, trunner, cfgs, spawn):
		CmdTestBase.__init__(self, cfg, trunner, cfgs, spawn)
		if trunner is None or self.coin not in self.networks:
			return
		if self.coin in ('btc', 'bch', 'ltc'):
			self.tx_fee     = {'btc':'0.0001', 'bch':'0.001', 'ltc':'0.01'}[self.coin]
			self.txbump_fee = {'btc':'123s', 'bch':'567s', 'ltc':'12345s'}[self.coin]

		self.unspent_data_file = joinpath('test', 'trash', 'unspent.json')
		self.spawn_env['MMGEN_BOGUS_UNSPENT_DATA'] = self.unspent_data_file

	@property
	def lbl_id(self):
		if not hasattr(self, '_lbl_id'):
			rpc = async_run(rpc_init(self.cfg, self.proto))
			self._lbl_id = ('account', 'label')['label_api' in rpc.caps]
		return self._lbl_id

	def _get_addrfile_checksum(self, display=False):
		addrfile = self.get_file_with_ext('addrs')
		from mmgen.addrlist import AddrList
		silence()
		chk = AddrList(self.cfg, self.proto, infile=addrfile).chksum
		end_silence()
		if self.cfg.verbose and display:
			msg(f'Checksum: {cyan(chk)}')
		return chk

	def walletgen_dfl_wallet(self, seed_len=None):
		return self.walletgen(seed_len=seed_len, gen_dfl_wallet=True)

	def subwalletgen_dfl_wallet(self, pf):
		return self.subwalletgen(wf='default')

	def export_seed_dfl_wallet(self, pf, out_fmt='seed'):
		return self.export_seed(wf=None, out_fmt=out_fmt, pf=pf)

	def addrgen_dfl_wallet(self, pf):
		return self.addrgen(wf=None, dfl_wallet=True)

	def txcreate_dfl_wallet(self, addrfile):
		return self.txcreate_common(sources=['15'])

	def txsign_dfl_wallet(self, txfile, pf='', save=True, has_label=False):
		return self.txsign(None, txfile, save=save, has_label=has_label, dfl_wallet=True)

	def passchg_dfl_wallet(self, pf):
		return self.passchg(wf=None, pf=pf, dfl_wallet=True)

	def walletchk_newpass_dfl_wallet(self, pf):
		return self.walletchk_newpass(wf=None, wcls=MMGenWallet, pf=pf, dfl_wallet=True)

	def delete_dfl_wallet(self, pf):
		self.write_to_tmpfile('del_dw_run', b'', binary=True)
		if self.cfg.no_dw_delete:
			return 'skip'
		for wf in [f for f in os.listdir(self.cfg.data_dir) if f[-6:]=='.mmdat']:
			os.unlink(joinpath(self.cfg.data_dir, wf))
		self.spawn(msg_only=True)
		self.have_dfl_wallet = False
		return 'ok'

	def walletgen(self, del_dw_run='dummy', seed_len=None, gen_dfl_wallet=False):
		self.write_to_tmpfile(pwfile, self.wpasswd+'\n')
		args = ['-p1']
		if not gen_dfl_wallet:
			args += ['-d', self.tmpdir]
		if seed_len:
			args += ['-l', str(seed_len)]
		t = self.spawn('mmgen-walletgen', self.testnet_opt + args + [self.usr_rand_arg], no_passthru_opts=True)
		t.license()
		t.usr_rand(self.usr_rand_chars)
		wcls = MMGenWallet
		t.passphrase_new('new '+wcls.desc, self.wpasswd)
		t.label()
		if not self.have_dfl_wallet and gen_dfl_wallet:
			t.expect('move it to the data directory? (Y/n): ', 'y')
			self.have_dfl_wallet = True
		t.written_to_file(capfirst(wcls.desc))
		return t

	def subwalletgen(self, wf):
		args = [self.usr_rand_arg, '-p1', '-d', self.tr.trash_dir, '-L', 'Label']
		if wf != 'default':
			args += [wf]
		t = self.spawn('mmgen-subwalletgen', self.testnet_opt + args + ['10s'], no_passthru_opts=True)
		t.license()
		wcls = MMGenWallet
		t.passphrase(wcls.desc, self.cfgs['1']['wpasswd'])
		t.expect(r'Generating subseed.*\D10S', regex=True)
		t.passphrase_new('new '+wcls.desc, 'foo')
		t.usr_rand(self.usr_rand_chars)
		fn = t.written_to_file(capfirst(wcls.desc))
		ext = get_extension(fn)
		assert ext, f'incorrect file extension: {ext}'
		return t

	def subwalletgen_mnemonic(self, wf):
		icls = get_wallet_cls(ext=get_extension(wf))
		ocls = get_wallet_cls('words')
		args = [self.usr_rand_arg, '-p1', '-d', self.tr.trash_dir, '-o', ocls.fmt_codes[0], wf, '3L']
		t = self.spawn('mmgen-subwalletgen', self.testnet_opt + args, no_passthru_opts=True)
		t.license()
		t.passphrase(icls.desc, self.cfgs['1']['wpasswd'])
		t.expect(r'Generating subseed.*\D3L', regex=True)
		fn = t.written_to_file(capfirst(ocls.desc))
		ext = get_extension(fn)
		assert ext == ocls.ext, f'incorrect file extension: {ext}'
		return t

	def passchg(self, wf, pf, label_action='cmdline', dfl_wallet=False, delete=False):
		silence()
		self.write_to_tmpfile(pwfile, get_data_from_file(self.cfg, pf))
		end_silence()
		add_args = {
			'cmdline': ['-d', self.tmpdir, '-L', 'Changed label (UTF-8) α'],
			'keep':    ['-d', self.tr.trash_dir, '--keep-label'],
			'user':    ['-d', self.tr.trash_dir]
		}[label_action]
		t = self.spawn(
			'mmgen-passchg',
			self.testnet_opt + add_args + [self.usr_rand_arg, '-p2'] + ([wf] if wf else []),
			no_passthru_opts = True)
		t.license()
		wcls = MMGenWallet
		t.passphrase(wcls.desc, self.cfgs['1']['wpasswd'], pwtype='old')
		t.expect_getend('Hash preset changed to ')
		t.passphrase(wcls.desc, self.wpasswd, pwtype='new') # reuse passphrase?
		t.expect('Repeat new passphrase: ', self.wpasswd+'\n')
		t.usr_rand(self.usr_rand_chars)
		if label_action == 'user':
			t.expect('Enter a wallet label.*: ', 'Interactive Label (UTF-8) α\n', regex=True)
		t.expect_getend(('Label changed to ', 'Reusing label ')[label_action=='keep'])
#		t.expect_getend('Key ID changed: ')
		if dfl_wallet:
			t.expect("Type uppercase 'YES' to confirm: ", 'YES\n')
			t.written_to_file('New wallet')
			t.expect('Securely deleting old wallet')
#			t.expect('Okay to WIPE 1 regular file ? (Yes/No)', 'Yes\n')
			t.expect('Wallet passphrase has changed')
			t.expect_getend('has been changed to ')
		else:
			t.written_to_file(capfirst(wcls.desc))
			t.expect('Securely delete .*: ', ('y' if delete else 'n'), regex=True)
			if delete:
				t.expect('Securely deleting')
		return t

	def passchg_keeplabel(self, wf, pf):
		return self.passchg(wf, pf, label_action='keep', delete=True)

	def passchg_usrlabel(self, wf, pf):
		return self.passchg(wf, pf, label_action='user')

	def walletchk_newpass(self, wf, pf, wcls=None, dfl_wallet=False):
		return self.walletchk(wf, wcls=wcls, dfl_wallet=dfl_wallet)

	def _write_fake_data_to_file(self, d):
		write_data_to_file(
				self.cfg,
				self.unspent_data_file,
				d,
				desc              = 'Unspent outputs',
				quiet             = True,
				ignore_opt_outdir = True)
		if not self.tr.quiet:
			sys.stderr.write(f'Fake transaction wallet data written to file {self.unspent_data_file!r}\n')

	def _create_fake_unspent_entry(
			self,
			coinaddr,
			al_id     = None,
			idx       = None,
			comment   = None,
			non_mmgen = False,
			segwit    = False):
		if 'S' not in self.proto.mmtypes:
			segwit = False
		if comment:
			comment = ' ' + comment
		k = coinaddr.addr_fmt
		if not segwit and k == 'p2sh':
			k = 'p2pkh'
		s_beg, s_end = {
			'p2pkh':  ('76a914', '88ac'),
			'p2sh':   ('a914', '87'),
			'bech32': (self.proto.witness_vernum_hex + '14', '')
		}[k]
		amt1, amt2 = {
			'btc': (10, 40),
			'bch': (10, 40),
			'ltc': (1000, 4000)
		}[self.coin]
		ret = {
			self.lbl_id: (
				f'{self.proto.base_coin.lower()}:{coinaddr}' if non_mmgen
				else f'{al_id}:{idx}{comment}'),
			'vout': int(getrandnum(4) % 8),
			'txid': getrandhex(32),
			'amount': self.proto.coin_amt('{}.{}'.format(
				amt1 + getrandnum(4) % amt2,
				getrandnum(4) % 100000000)),
			'address': coinaddr,
			'spendable': False,
			'scriptPubKey': f'{s_beg}{coinaddr.bytes.hex()}{s_end}',
			'confirmations': getrandnum(3) // 20 # max: 838860 (6 digits)
		}
		return ret

	def _create_fake_unspent_data(
			self,
			adata,
			tx_data,
			non_mmgen_input            = '',
			non_mmgen_input_compressed = True):

		out = []
		for d in tx_data.values():
			al = adata.addrlist(al_id=d['al_id'])
			for n, (idx, coinaddr) in enumerate(al.addrpairs()):
				comment = get_comment(do_shuffle=not self.cfg.test_suite_deterministic)
				out.append(self._create_fake_unspent_entry(
					coinaddr, d['al_id'], idx, comment, segwit=d['segwit']))
				if n == 0:  # create a duplicate address. This means addrs_per_wallet += 1
					out.append(self._create_fake_unspent_entry(
						coinaddr, d['al_id'], idx, comment, segwit=d['segwit']))

		if non_mmgen_input:
			from mmgen.key import PrivKey
			privkey = PrivKey(
				self.proto,
				getrand(32),
				compressed  = non_mmgen_input_compressed,
				pubkey_type = 'std')
			from mmgen.addrgen import KeyGenerator, AddrGenerator
			rand_coinaddr = AddrGenerator(
				self.cfg,
				self.proto,
				('legacy', 'compressed')[non_mmgen_input_compressed]
				).to_addr(KeyGenerator(self.cfg, self.proto, 'std').gen_data(privkey))
			of = joinpath(self.cfgs[non_mmgen_input]['tmpdir'], non_mmgen_fn)
			write_data_to_file(
				cfg               = self.cfg,
				outfile           = of,
				data              = privkey.wif + '\n',
				desc              = f'compressed {self.proto.name} key',
				quiet             = True,
				ignore_opt_outdir = True)
			out.append(self._create_fake_unspent_entry(rand_coinaddr, non_mmgen=True, segwit=False))

		return out

	def _create_tx_data(self, sources, addrs_per_wallet=addrs_per_wallet):
		from mmgen.addrlist import AddrList, AddrIdxList
		from mmgen.addrdata import AddrData
		tx_data, ad = {}, AddrData(self.proto)
		for s in sources:
			addrfile = get_file_with_ext(self.cfgs[s]['tmpdir'], 'addrs')
			al = AddrList(self.cfg, self.proto, infile=addrfile)
			ad.add(al)
			aix = AddrIdxList(fmt_str=self.cfgs[s]['addr_idx_list'])
			if len(aix) != addrs_per_wallet:
				die('TestSuiteFatalException', f'Address index list length != {addrs_per_wallet}: {repr(aix)}')
			tx_data[s] = {
				'addrfile': addrfile,
				'chk': al.chksum,
				'al_id': al.al_id,
				'addr_idxs': aix[-2:],
				'segwit': self.cfgs[s]['segwit']
			}
		return ad, tx_data

	def _make_txcreate_cmdline(self, tx_data):
		from mmgen.key import PrivKey
		privkey = PrivKey(self.proto, getrand(32), compressed=True, pubkey_type='std')
		t = ('compressed', 'segwit')['S' in self.proto.mmtypes]
		from mmgen.addrgen import KeyGenerator, AddrGenerator
		rand_coinaddr = AddrGenerator(self.cfg, self.proto, t).to_addr(
			KeyGenerator(self.cfg, self.proto, 'std').gen_data(privkey)
		)

		# total of two outputs must be < 10 BTC (<1000 LTC)
		mods = {
			'btc': (6, 4),
			'bch': (6, 4),
			'ltc': (600, 400)
		}[self.coin]
		for k in self.cfgs:
			self.cfgs[k]['amts'] = [None, None]
			for idx, mod in enumerate(mods):
				self.cfgs[k]['amts'][idx] = '{}.{}'.format(
					getrandnum(4) % mod,
					str(getrandnum(4))[:5])

		cmd_args = ['--outdir='+self.tmpdir]
		for num in tx_data:
			s = tx_data[num]
			cmd_args += [
				'{}:{},{}'.format(
					s['al_id'],
					s['addr_idxs'][0],
					self.cfgs[num]['amts'][0])]
			# + one change address and one BTC address
			if num is list(tx_data.keys())[-1]:
				cmd_args += [
					'{}:{}'.format(
						s['al_id'],
						s['addr_idxs'][1])]
				cmd_args += [
					'{},{}'.format(
						rand_coinaddr,
						self.cfgs[num]['amts'][1])]

		return cmd_args + [tx_data[num]['addrfile'] for num in tx_data]

	def txcreate_common(
			self,
			sources                    = ['1'],
			non_mmgen_input            = '',
			do_label                   = False,
			txdo_args                  = [],
			add_args                   = [],
			view                       = 'n',
			addrs_per_wallet           = addrs_per_wallet,
			non_mmgen_input_compressed = True,
			cmdline_inputs             = False,
			tweaks                     = []):

		if not self.tr.quiet:
			sys.stderr.write(green('Generating fake tracking wallet info\n'))

		silence()
		ad, tx_data = self._create_tx_data(sources, addrs_per_wallet)
		dfake = self._create_fake_unspent_data(ad, tx_data, non_mmgen_input, non_mmgen_input_compressed)
		import json
		from mmgen.rpc.util import json_encoder
		self._write_fake_data_to_file(json.dumps(dfake, cls=json_encoder))
		cmd_args = self._make_txcreate_cmdline(tx_data)

		if cmdline_inputs:
			from mmgen.tw.shared import TwLabel
			cmd_args = [
				'--inputs={},{},{},{},{},{}'.format(
					TwLabel(self.proto, dfake[0][self.lbl_id]).mmid, dfake[1]['address'],
					TwLabel(self.proto, dfake[2][self.lbl_id]).mmid, dfake[3]['address'],
					TwLabel(self.proto, dfake[4][self.lbl_id]).mmid, dfake[5]['address']
				),
				f'--outdir={self.tr.trash_dir}'
			] + cmd_args[1:]

		end_silence()

		if not self.tr.quiet:
			sys.stderr.write('\n')

		t = self.spawn(
			'mmgen-'+('txcreate', 'txdo')[bool(txdo_args)],
			(['--no-rbf'], [])[self.proto.cap('rbf')] +
			['-f', self.tx_fee, '-B'] + add_args + cmd_args + txdo_args)

		if t.expect([('Get', 'Unsigned transac')[cmdline_inputs], r'Unable to connect to \S+'], regex=True) == 1:
			die('TestSuiteException', '\n'+t.p.after)

		if cmdline_inputs:
			t.written_to_file('tion')
			return t

		t.license()

		for num in tx_data:
			t.expect_getend('ting address data from file ')
			chk=t.expect_getend(r'Checksum for address data .*?: ', regex=True)
			verify_checksum_or_exit(tx_data[num]['chk'], chk)

		# not in tracking wallet warning, (1 + num sources) times
		for num in range(len(tx_data) + 1):
			t.expect('Continue anyway? (y/N): ', 'y')

		outputs_list = [(addrs_per_wallet+1)*i + 1 for i in range(len(tx_data))]
		if non_mmgen_input:
			outputs_list.append(len(tx_data)*(addrs_per_wallet+1) + 1)

		self.txcreate_ui_common(t,
			menu        = (['M'], ['M', 'D', 'D', 'D', 'D', 'm', 'o'])[self.test_name=='txcreate'],
			inputs      = ' '.join(map(str, outputs_list)),
			add_comment = ('', tx_comment_lat_cyr_gr)[do_label],
			view        = view,
			tweaks      = tweaks)

		if txdo_args and add_args: # txdo4
			t.do_decrypt_ka_data(pw=self.cfgs['14']['kapasswd'])

		return t

	def txcreate(self, addrfile):
		return self.txcreate_common(sources=['1'], add_args=['--vsize-adj=1.01'])

	def txbump(self, txfile, prepend_args=[], seed_args=[]):
		if not self.proto.cap('rbf'):
			msg(gray('Skipping RBF'))
			return 'skip'
		args = prepend_args + ['--quiet', '--outdir='+self.tmpdir, txfile] + seed_args
		t = self.spawn('mmgen-txbump', args)
		if seed_args:
			t.do_decrypt_ka_data(pw=self.cfgs['14']['kapasswd'])
		t.expect('deduct the fee from (Hit ENTER for the change output): ', '1\n')
		# Fee must be > tx_fee + network relay fee (currently 0.00001)
		t.expect('OK? (Y/n): ', '\n')
		t.expect('Enter transaction fee: ', self.txbump_fee+'\n')
		t.expect('OK? (Y/n): ', '\n')
		if seed_args: # sign and send
			t.do_comment(False, has_label=True)
			for cnum, wcls in (('1', IncogWallet), ('3', MMGenWallet), ('4', MMGenWallet)):
				t.passphrase(wcls.desc, self.cfgs[cnum]['wpasswd'])
			self._do_confirm_send(t, quiet=not self.cfg.debug, confirm_send=True)
			if self.cfg.debug:
				t.written_to_file('Transaction')
		else:
			t.do_comment(False)
			t.expect('Save fee-bumped transaction? (y/N): ', 'y')
			t.written_to_file('Fee-bumped transaction')
		os.unlink(txfile) # our tx file replaces the original
		cmd = 'touch ' + joinpath(self.tmpdir, 'txbump')
		os.system(cmd)
		return t

	def txsend(self, sigfile, extra_opts=[]):
		t = self.spawn('mmgen-txsend', extra_opts + ['-d', self.tmpdir, sigfile], no_passthru_opts=['coin'])
		self.txsend_ui_common(t, view='t', add_comment='')
		return t

	def txdo(self, addrfile, wallet):
		t = self.txcreate_common(sources=['1'], txdo_args=[wallet])
		self.txsign_ui_common(t, view='n', do_passwd=True)
		self.txsend_ui_common(t)
		return t

	def _walletconv_export(self, wf, uargs=[], out_fmt='w', pf=None):
		opts = ['-d', self.tmpdir, '-o', out_fmt] + uargs + \
			([], [wf])[bool(wf)] + ([], ['-P', pf])[bool(pf)]
		t = self.spawn('mmgen-walletconv', self.testnet_opt + opts, no_passthru_opts=True)
		t.license()

		if not pf:
			icls = get_wallet_cls(ext=get_extension(wf))
			t.passphrase(icls.desc, self.wpasswd)

		ocls = get_wallet_cls(fmt_code=out_fmt)

		if ocls.enc and ocls.type != 'brain':
			t.passphrase_new('new '+ocls.desc, self.wpasswd)
			t.usr_rand(self.usr_rand_chars)

		if ocls.type.startswith('incog'):
			m = 'Encrypting random data from your operating system with ephemeral key'
			t.expect(m)
			t.expect(m)
			incog_id = t.expect_getend('New Incog Wallet ID: ')
			t.expect(m)

		if ocls.type == 'incog_hidden':
			self.write_to_tmpfile(incog_id_fn, incog_id)
			t.hincog_create(hincog_bytes)
		elif ocls.type == 'mmgen':
			t.label()

		return t.written_to_file(capfirst(ocls.desc)), t

	def export_seed(self, wf, out_fmt='seed', pf=None):
		f, t = self._walletconv_export(wf, out_fmt=out_fmt, pf=pf)
		silence()
		wcls = get_wallet_cls(fmt_code=out_fmt)
		msg('==> {}: {}'.format(
			wcls.desc,
			cyan(get_data_from_file(self.cfg, f, desc=wcls.desc))
		))
		end_silence()
		return t

	def export_hex(self, wf, out_fmt='mmhex', pf=None):
		return self.export_seed(wf, out_fmt=out_fmt, pf=pf)

	def export_mnemonic(self, wf):
		return self.export_seed(wf, out_fmt='words')

	def export_bip39(self, wf):
		return self.export_seed(wf, out_fmt='bip39')

	def export_incog(self, wf, out_fmt='i', add_args=[]):
		uargs = ['-p1', self.usr_rand_arg] + add_args
		_, t = self._walletconv_export(wf, out_fmt=out_fmt, uargs=uargs)
		return t

	def export_incog_hex(self, wf):
		return self.export_incog(wf, out_fmt='xi')

	# TODO: make outdir and hidden incog compatible (ignore --outdir and warn user?)
	def export_incog_hidden(self, wf):
		rf = joinpath(self.tmpdir, hincog_fn)
		add_args = ['-J', f'{rf},{hincog_offset}']
		return self.export_incog(wf, out_fmt='hi', add_args=add_args)

	def addrgen_seed(self, wf, _, in_fmt='seed'):
		wcls = get_wallet_cls(fmt_code=in_fmt)
		stdout = wcls.type == 'seed' # capture output to screen once
		t = self.spawn(
			'mmgen-addrgen',
			(['-S'] if stdout else [])
			+ self.segwit_arg
			+ ['-i' + in_fmt, '-d', self.tmpdir, wf, self.addr_idx_list],
			exit_val = None if stdout else 1)
		t.license()
		t.expect_getend(f'Valid {wcls.desc} for Seed ID ')
		vmsg('Comparing generated checksum with checksum from previous address file')
		verify_checksum_or_exit(
			self._get_addrfile_checksum(),
			t.expect_getend(r'Checksum for address data .*?: ', regex=True))
		if not stdout:
			t.no_overwrite()
		return t

	def addrgen_hex(self, wf, _, in_fmt='mmhex'):
		return self.addrgen_seed(wf, _, in_fmt=in_fmt)

	def addrgen_mnemonic(self, wf, _):
		return self.addrgen_seed(wf, _, in_fmt='words')

	def addrgen_incog(self, wf=[], _='', in_fmt='i', args=[]):
		t = self.spawn(
			'mmgen-addrgen',
			args
			+ self.segwit_arg
			+ ['-i'+in_fmt, '-d', self.tmpdir]
			+ ([wf] if wf else [])
			+ [self.addr_idx_list],
			exit_val = 1)
		t.license()
		t.expect_getend('Incog Wallet ID: ')
		wcls = get_wallet_cls(fmt_code=in_fmt)
		t.hash_preset(wcls.desc, '1')
		t.passphrase(rf'{wcls.desc} \w{{8}}', self.wpasswd)
		vmsg('Comparing generated checksum with checksum from address file')
		chk = t.expect_getend(r'Checksum for address data .*?: ', regex=True)
		verify_checksum_or_exit(self._get_addrfile_checksum(), chk)
		t.no_overwrite()
		return t

	def addrgen_incog_hex(self, wf, _):
		return self.addrgen_incog(wf, '', in_fmt='xi')

	def addrgen_incog_hidden(self, wf, _):
		rf = joinpath(self.tmpdir, hincog_fn)
		return self.addrgen_incog([], '', in_fmt='hi',
			args=['-H', f'{rf},{hincog_offset}', '-l', str(hincog_seedlen)])

	def txsign_keyaddr(self, keyaddr_file, txfile):
		t = self.spawn('mmgen-txsign',
			['-d', self.tmpdir, '-p1', '-M', keyaddr_file, txfile], no_passthru_opts=['coin'])
		t.license()
		t.view_tx('n')
		t.do_decrypt_ka_data(pw=self.kapasswd)
		self.txsign_end(t)
		return t

	def txcreate_ni(self, addrfile):
		return self.txcreate_common(sources=['1'], cmdline_inputs=True, add_args=['--yes'])

	def walletgen2(self, del_dw_run='dummy'):
		return self.walletgen(seed_len=128)

	def addrgen2(self, wf):
		return self.addrgen(wf)

	def txcreate2(self, addrfile):
		return self.txcreate_common(sources=['2'])

	def txsign2(self, wf1, txf1, wf2, txf2):
		t = self.spawn('mmgen-txsign',
			['-d', self.tmpdir, txf1, wf1, txf2, wf2], no_passthru_opts=['coin'])
		t.license()
		for cnum, wf in (('1', wf1), ('2', wf2)):
			wcls = get_wallet_cls(ext=get_extension(wf))
			t.view_tx('n')
			t.passphrase(wcls.desc, self.cfgs[cnum]['wpasswd'])
			self.txsign_end(t, cnum)
		return t

	def export_mnemonic2(self, wf):
		return self.export_mnemonic(wf)

	def walletgen3(self, del_dw_run='dummy'):
		return self.walletgen()

	def addrgen3(self, wf):
		return self.addrgen(wf)

	def txcreate3(self, addrfile1, addrfile2):
		return self.txcreate_common(sources=['1', '3'])

	def txsign3(self, wf1, wf2, txf2):
		t = self.spawn('mmgen-txsign', ['-d', self.tmpdir, wf1, wf2, txf2], no_passthru_opts=['coin'])
		t.license()
		t.view_tx('n')
		for cnum, wf in (('1', wf1), ('3', wf2)):
			wcls = get_wallet_cls(ext=get_extension(wf))
			t.passphrase(wcls.desc, self.cfgs[cnum]['wpasswd'])
		self.txsign_end(t)
		return t

	walletgen14 = walletgen
	addrgen14 = CmdTestShared.addrgen
	keyaddrgen14 = CmdTestShared.keyaddrgen

	def walletgen4(self, del_dw_run='dummy'):
		bwf = joinpath(self.tmpdir, self.bw_filename)
		make_brainwallet_file(self.cfg, bwf)
		seed_len = str(self.seed_len)
		args = ['-d', self.tmpdir, '-p1', self.usr_rand_arg, '-l'+seed_len, '-ibw']
		t = self.spawn('mmgen-walletconv', self.testnet_opt + args + [bwf], no_passthru_opts=True)
		t.license()
		wcls = MMGenWallet
		t.passphrase_new('new ' +wcls.desc, self.wpasswd)
		t.usr_rand(self.usr_rand_chars)
		t.label()
		t.written_to_file(capfirst(wcls.desc))
		return t

	def addrgen4(self, wf):
		return self.addrgen(wf)

	def txcreate4(self, f1, f2, f3, f4, f5, f6):
		return self.txcreate_common(
			sources         = ['1', '2', '3', '4', '14'],
			non_mmgen_input = '4',
			do_label        = True,
			view            = 'y',
			tweaks          = ['confirm_non_mmgen'])

	def txsign4(self, f1, f2, f3, f4, f5, f6):
		non_mm_file = joinpath(self.tmpdir, non_mmgen_fn)
		add_args = [
			'-d', self.tmpdir,
			'-i', 'brain',
			'-b' + self.bw_params,
			'-p1',
			'--keys-from-file=' + non_mm_file,
			'--mmgen-keys-from-file=' + f6,
			f1, f2, f3, f4, f5]
		t = self.spawn('mmgen-txsign', add_args, no_passthru_opts=['coin'])
		t.license()
		t.view_tx('t')
		t.do_decrypt_ka_data(pw=self.cfgs['14']['kapasswd'])

		for cnum, wcls in (('1', IncogWallet), ('3', MMGenWallet)):
			t.passphrase(wcls.desc, self.cfgs[cnum]['wpasswd'])

		self.txsign_end(t, has_label=True)
		return t

	def txdo4(self, f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11, f12):
		non_mm_file = joinpath(self.tmpdir, non_mmgen_fn)
		add_args = [
			'-d', self.tmpdir,
			'-i', 'brain',
			'-b'+self.bw_params,
			'-p1',
			'--keys-from-file=' + non_mm_file,
			'--mmgen-keys-from-file=' + f12]
		self.get_file_with_ext('sigtx', delete_all=True) # delete tx signed by txsign4
		t = self.txcreate_common(
			sources         = ['1', '2', '3', '4', '14'],
			non_mmgen_input = '4',
			do_label        = True,
			txdo_args       = [f7, f8, f9, f10],
			add_args        = add_args)

		for cnum, wcls in (('1', IncogWallet), ('3', MMGenWallet)):
			t.passphrase(wcls.desc, self.cfgs[cnum]['wpasswd'])

		self.txsign_ui_common(t)
		self.txsend_ui_common(t)

		cmd = 'touch ' + joinpath(self.tmpdir, 'txdo')
		os.system(cmd)
		return t

	def txbump4(self, f1, f2, f3, f4, f5, f6, f7, f8, f9): # f7:txfile, f9:'txdo'
		non_mm_file = joinpath(self.tmpdir, non_mmgen_fn)
		return self.txbump(
			f7,
			prepend_args = ['-p1', '-k', non_mm_file, '-M', f1],
			seed_args    = [f2, f3, f4, f6, f8])

	def walletgen5(self, del_dw_run='dummy'):
		return self.walletgen()

	def addrgen5(self, wf):
		return self.addrgen(wf)

	def txcreate5(self, addrfile):
		return self.txcreate_common(
			sources                    = ['20'],
			non_mmgen_input            = '20',
			non_mmgen_input_compressed = False,
			tweaks                     = ['confirm_non_mmgen'])

	def txsign5(self, wf, txf, bad_vsize=True, add_args=[]):
		non_mm_file = joinpath(self.tmpdir, non_mmgen_fn)
		t = self.spawn(
			'mmgen-txsign',
			add_args + ['-d', self.tmpdir, '-k', non_mm_file, txf, wf],
			no_passthru_opts = ['coin'],
			exit_val = 2 if bad_vsize else None)
		t.license()
		t.view_tx('n')
		wcls = get_wallet_cls(ext=get_extension(wf))
		t.passphrase(wcls.desc, self.cfgs['20']['wpasswd'])
		if bad_vsize:
			t.expect('Estimated transaction vsize')
			t.expect('1 transaction could not be signed')
		else:
			t.do_comment(False)
			t.expect('Save signed transaction? (Y/n): ', 'y')
		return t

	def walletgen6(self, del_dw_run='dummy'):
		return self.walletgen()

	def addrgen6(self, wf):
		return self.addrgen(wf)

	def txcreate6(self, addrfile):
		return self.txcreate_common(
			sources                    = ['21'],
			non_mmgen_input            = '21',
			non_mmgen_input_compressed = False,
			add_args                   = ['--vsize-adj=1.08'],
			tweaks                     = ['confirm_non_mmgen'])

	def txsign6(self, wf, txf):
		return self.txsign5(wf, txf, bad_vsize=False, add_args=['--vsize-adj=1.08'])



#!/usr/bin/env python3

"""
test.daemontest_d.msg: message signing unit tests for the MMGen suite
"""

import os

from mmgen.util import msg, pumsg, suf
from mmgen.protocol import CoinProtocol
from mmgen.msg import NewMsg, UnsignedMsg, SignedMsg, SignedOnlineMsg, ExportedMsgSigs
from mmgen.addr import MMGenID

from ..include.common import cfg, silence, end_silence, restart_test_daemons, stop_test_daemons

def get_obj(coin, network, msghash_type):

	if coin == 'bch':
		addrlists = 'DEADBEEF:C:1-20 98831F3A:C:8,2 A091ABAA:L:111 A091ABAA:C:1'
	elif coin == 'eth':
		addrlists = 'DEADBEEF:E:1-20 98831F3A:E:8,2 A091ABAA:E:111'
	else:
		# A091ABAA = 98831F3A:5S
		addrlists = 'DEADBEEF:C:1-20 98831F3A:B:8,2 A091ABAA:S:10-11 A091ABAA:111 A091ABAA:C:1'

	return NewMsg(
		cfg       = cfg,
		coin      = coin,
		network   = network,
		message   = '08/Jun/2021 Bitcoin Law Enacted by El Salvador Legislative Assembly',
		addrlists = addrlists,
		msghash_type = msghash_type)

def print_total(n):
	msg(f'{n} signature{suf(n)} verified')

async def run_test(network_id, chksum, msghash_type='raw'):

	coin, network = CoinProtocol.Base.parse_network_id(network_id)

	if not cfg.verbose:
		silence()

	m = get_obj(coin, network, msghash_type)

	if m.proto.sign_mode == 'daemon':
		restart_test_daemons(network_id)

	pumsg('\nTesting data creation:\n')

	tmpdir = os.path.join('test', 'trash2')

	os.makedirs(tmpdir, exist_ok=True)

	assert m.chksum.upper() == chksum, f'{m.chksum.upper()} != {chksum}'

	m.write_to_file(
		outdir        = tmpdir,
		ask_overwrite = False)

	pumsg('\nTesting signing:\n')

	m = UnsignedMsg(cfg, infile = os.path.join(tmpdir, get_obj(coin, network, msghash_type).filename))
	await m.sign(wallet_files=['test/ref/98831F3A.mmwords'])

	m = SignedMsg(cfg, data=m.__dict__)
	m.write_to_file(
		outdir        = tmpdir,
		ask_overwrite = False)

	pumsg('\nTesting display:\n')

	m = SignedOnlineMsg(cfg, infile = os.path.join(tmpdir, get_obj(coin, network, msghash_type).signed_filename))

	msg(m.format())

	single_addr = 'A091ABAA:E:111' if m.proto.base_proto == 'Ethereum' else 'A091ABAA:111'
	single_addr_coin = m.sigs[MMGenID(m.proto, single_addr)]['addr']

	pumsg('\nTesting single address display:\n')
	msg(m.format(single_addr))

	pumsg('\nTesting verification:\n')
	print_total(await m.verify())

	pumsg('\nTesting single address verification:\n')
	print_total(await m.verify(addr=single_addr))

	pumsg('\nTesting JSON dump for export:\n')
	msg(m.get_json_for_export())

	pumsg('\nTesting single address JSON dump for export:\n')
	msg(m.get_json_for_export(addr=single_addr))

	from mmgen.fileutil import write_data_to_file
	exported_sigs = os.path.join(tmpdir, 'signatures.json')
	write_data_to_file(
		cfg     = cfg,
		outfile = exported_sigs,
		data    = m.get_json_for_export(),
		desc    = 'signature data',
		ask_overwrite = False)

	m = ExportedMsgSigs(cfg, infile=exported_sigs)

	pumsg('\nTesting verification (exported data):\n')
	print_total(await m.verify())

	pumsg('\nTesting single address verification (exported data):\n')
	print_total(await m.verify(addr=single_addr_coin))

	pumsg('\nTesting display (exported data):\n')
	msg(m.format())

	pumsg('\nTesting single address display (exported data):\n')
	msg(m.format(single_addr_coin))

	if m.proto.sign_mode == 'daemon':
		stop_test_daemons(network_id, remove_datadir=True)

	msg('\n')

	if not cfg.verbose:
		end_silence()

	return True

class unit_tests:

	altcoin_deps = ('ltc', 'bch', 'eth', 'eth_raw')

	def btc(self, name, ut, desc='Bitcoin mainnet'):
		return run_test('btc', 'AA0DB5')

	def btc_tn(self, name, ut, desc='Bitcoin testnet'):
		return run_test('btc_tn', 'A88E1D')

	def btc_rt(self, name, ut, desc='Bitcoin regtest'):
		return run_test('btc_rt', '578018')

	def ltc(self, name, ut, desc='Litecoin mainnet'):
		return run_test('ltc', 'BA7549')

	def bch(self, name, ut, desc='Bitcoin Cash mainnet'):
		return run_test('bch', '1B8065')

	def eth(self, name, ut, desc='Ethereum mainnet'):
		return run_test('eth', '35BAD9', msghash_type='eth_sign')

	def eth_raw(self, name, ut, desc='Ethereum mainnet (raw message)'):
		return run_test('eth', '9D900C')
